from mage_ai.data_preparation.repo_manager import get_repo_config
from mage_ai.orchestration.queue.config import QueueConfig, QueueType


class QueueFactory:
    queue = None

    @classmethod
    def get_queue(self):
        if self.queue is not None:
            return self.queue
        queue_config = QueueConfig.load(config=get_repo_config().queue_config)
        if queue_config.queue_type == QueueType.CELERY:
            from mage_ai.orchestration.queue.celery_queue import CeleryQueue
            self.queue = CeleryQueue(queue_config)
        else:
            from mage_ai.orchestration.queue.process_queue import ProcessQueue
            self.queue = ProcessQueue(queue_config)
        return self.queue



import enum
import re
from datetime import datetime
from typing import Dict, List, Union

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    String,
)
from sqlalchemy.orm import relationship, validates

from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db import db_connection, safe_db_query
from mage_ai.orchestration.db.errors import ValidationError
from mage_ai.orchestration.db.models.base import BaseModel


class User(BaseModel):
    avatar = Column(String(255), default=None)
    email = Column(String(255), default=None, index=True, unique=True)
    first_name = Column(String(255), default=None)
    last_name = Column(String(255), default=None)
    _owner = Column('owner', Boolean, default=False)
    password_hash = Column(String(255), default=None)
    password_salt = Column(String(255), default=None)
    roles = Column(Integer, default=None)
    username = Column(String(255), default=None, index=True, unique=True)
    preferences = Column(JSON, default=None)

    oauth2_applications = relationship('Oauth2Application', back_populates='user')
    oauth2_access_tokens = relationship('Oauth2AccessToken', back_populates='user')
    # roles_new is used for the new authentication system to define permissions at
    # an entity level
    roles_new = relationship('Role', secondary='user_role', back_populates='users')

    @validates('email')
    def validate_email(self, key, value):
        if not value:
            raise ValidationError('Email address cannot be blank.', metadata=dict(
                key=key,
                value=value,
            ))
        else:
            existing_email = User.query.filter(
                User.email == value
            ).one_or_none()
            if self.email != value and existing_email is not None:
                raise ValidationError(
                    'Email address is already in use. Please choose a different one.',
                    metadata=dict(
                        key=key,
                        value=value,
                    )
                )

            regex = re.compile(r"([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\"([]!#-[^-~ \t]|(\\[\t -~]))+\")@([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\[[\t -Z^-~]*])")  # noqa: E501
            if not re.fullmatch(regex, value):
                raise ValidationError('Email address is invalid.', metadata=dict(
                    key=key,
                    value=value,
                ))

        return value

    @property
    def project_access(self) -> int:
        return self.get_access(Permission.Entity.PROJECT, get_repo_path())

    def get_access(
        self,
        entity: Union['Permission.Entity', None] = None,
        entity_id: Union[str, None] = None,
    ) -> int:
        '''
        If entity is None, we will go through all of the user's permissions and
        get the "highest" permission regardless of entity type. This should only be
        used for resources that are not entity dependent.

        Otherwise, search for permissions for the specified entity and entity_id, and
        return the access of the user for that entity.
        '''
        access = 0
        for role in self.roles_new:
            access = access | role.get_access(entity, entity_id)
        return access

    @property
    def roles_display(self) -> str:
        if self.owner:
            return 'Owner'
        elif self.roles:
            if self.roles & 1 != 0:
                return 'Admin'
            elif self.roles & 2 != 0:
                return 'Editor'
            elif self.roles & 4 != 0:
                return 'Viewer'

    @property
    def owner(self) -> bool:
        access = self.project_access if self.roles_new else 0
        return self._owner or access & Permission.Access.OWNER != 0

    @property
    def is_admin(self) -> bool:
        if self.roles_new:
            access = self.project_access
            return access & \
                (Permission.Access.OWNER | Permission.Access.ADMIN) == Permission.Access.ADMIN
        if not self.owner and self.roles:
            return self.roles & 1 != 0

        return False

    @property
    def git_settings(self) -> Union[Dict, None]:
        preferences = self.preferences or dict()
        return preferences.get(get_repo_path(), {}).get('git_settings')

    @classmethod
    @safe_db_query
    def batch_update_user_roles(self):
        for user in User.query.all():
            roles_new = []
            if user._owner:
                roles_new = [Role.get_role('Owner')]
            elif user.roles and user.roles & 1 != 0:
                roles_new = [Role.get_role('Admin')]
            elif user.roles and user.roles & 2 != 0:
                roles_new = [Role.get_role('Editor')]
            elif user.roles and user.roles & 4 != 0:
                roles_new = [Role.get_role('Viewer')]
            user.roles_new = roles_new
        db_connection.session.commit()


class Role(BaseModel):
    name = Column(String(255))
    permissions = relationship('Permission', back_populates='role')
    users = relationship('User', secondary='user_role', back_populates='roles_new')

    @classmethod
    @safe_db_query
    def create_default_roles(self):
        Permission.create_default_global_permissions()
        mapping = {
            'Owner': Permission.Access.OWNER,
            'Admin': Permission.Access.ADMIN,
            'Editor': Permission.Access.EDITOR,
            'Viewer': Permission.Access.VIEWER,
        }
        for name, access in mapping.items():
            role = self.query.filter(self.name == name).first()
            if not role:
                self.create(
                    name=name,
                    permissions=[
                        Permission.query.filter(
                            Permission.entity == Permission.Entity.GLOBAL,
                            Permission.access == access,
                        ).first()
                    ],
                    commit=False,
                )
        db_connection.session.commit()

    @classmethod
    @safe_db_query
    def get_role(self, name) -> 'Role':
        return Role.query.filter(Role.name == name).first()

    def get_access(
        self,
        entity: Union['Permission.Entity', None] = None,
        entity_id: Union[str, None] = None,
    ) -> int:
        permissions = []
        if entity is None:
            permissions.extend(self.permissions)
        else:
            entity_permissions = list(filter(
                lambda perm: perm.entity == entity and
                (entity_id is None or perm.entity_id == entity_id),
                self.permissions,
            ))
            if entity_permissions:
                permissions.extend(entity_permissions)

        access = 0
        if permissions:
            for permission in permissions:
                access = access | permission.access
            return access
        else:
            # TODO: Handle permissions with different entity types better.
            return self.get_parent_access(entity)

    def get_parent_access(self, entity) -> int:
        '''
        This method is used when a role does not have a permission for a specified entity. Then,
        we will go up the entity chain to see if there are permissions for parent entities.
        '''
        if entity == Permission.Entity.PIPELINE:
            return self.get_access(Permission.Entity.PROJECT, get_repo_path())
        elif entity == Permission.Entity.PROJECT:
            return self.get_access(Permission.Entity.GLOBAL)
        else:
            return 0


class UserRole(BaseModel):
    user_id = Column(Integer, ForeignKey('user.id'))
    role_id = Column(Integer, ForeignKey('role.id'))


class Permission(BaseModel):
    class Entity(str, enum.Enum):
        GLOBAL = 'global'
        PROJECT = 'project'
        PIPELINE = 'pipeline'

    class Access(int, enum.Enum):
        OWNER = 1
        ADMIN = 2
        EDITOR = 4
        VIEWER = 8

    entity_id = Column(String(255))
    entity = Column(Enum(Entity), default=Entity.GLOBAL)
    # 1 = owner
    # 2 = admin
    # 4 = edit
    # 8 = view
    access = Column(Integer, default=None)
    role_id = Column(Integer, ForeignKey('role.id'))

    role = relationship(Role, back_populates='permissions')

    @classmethod
    @safe_db_query
    def create_default_global_permissions(self) -> List['Permission']:
        permissions = self.query.filter(self.entity == Permission.Entity.GLOBAL).all()
        if len(permissions) == 0:
            for access in [a.value for a in Permission.Access]:
                permissions.append(
                    self.create(
                        entity=Permission.Entity.GLOBAL,
                        access=access,
                        commit=False,
                    )
                )
            db_connection.session.commit()
        return permissions


class Oauth2Application(BaseModel):
    class AuthorizationGrantType(str, enum.Enum):
        AUTHORIZATION_CODE = 'authorization-code'
        CLIENT_CREDENTIALS = 'client-credentials'

    class ClientType(str, enum.Enum):
        PRIVATE = 'private'
        PUBLIC = 'public'

    authorization_grant_type = Column(
        Enum(AuthorizationGrantType),
        default=AuthorizationGrantType.AUTHORIZATION_CODE,
    )
    client_id = Column(String(255), index=True, unique=True)
    client_type = Column(Enum(ClientType), default=ClientType.PRIVATE)
    name = Column(String(255))
    redirect_uris = Column(String(255), default=None)
    user = relationship(User, back_populates='oauth2_applications')
    user_id = Column(Integer, ForeignKey('user.id'))

    oauth2_access_tokens = relationship('Oauth2AccessToken', back_populates='oauth2_application')

    @classmethod
    @safe_db_query
    def query_client(self, api_key: str):
        return self.query.filter(
            Oauth2Application.client_id == api_key,
        ).first()


class Oauth2AccessToken(BaseModel):
    expires = Column(DateTime(timezone=True))
    oauth2_application = relationship(Oauth2Application, back_populates='oauth2_access_tokens')
    oauth2_application_id = Column(Integer, ForeignKey('oauth2_application.id'))
    token = Column(String(255), index=True, unique=True)
    user = relationship(User, back_populates='oauth2_access_tokens')
    user_id = Column(Integer, ForeignKey('user.id'))

    def is_valid(self) -> bool:
        return self.token and \
            self.expires and \
            self.expires >= datetime.utcnow().replace(tzinfo=self.expires.tzinfo)



from dataclasses import dataclass
from kafka import KafkaProducer
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sinks.base import BaseSink
from enum import Enum
from typing import Dict, List
import json
import time


class SecurityProtocol(str, Enum):
    SASL_SSL = 'SASL_SSL'
    SSL = 'SSL'


@dataclass
class SASLConfig:
    mechanism: str = 'PLAIN'
    username: str = None
    password: str = None


@dataclass
class SSLConfig:
    cafile: str = None
    certfile: str = None
    keyfile: str = None
    password: str = None
    check_hostname: bool = False


@dataclass
class KafkaConfig(BaseConfig):
    bootstrap_server: str
    topic: str
    api_version: str = '0.10.2'
    security_protocol: SecurityProtocol = None
    ssl_config: SSLConfig = None
    sasl_config: SASLConfig = None

    @classmethod
    def parse_config(self, config: Dict) -> Dict:
        ssl_config = config.get('ssl_config')
        sasl_config = config.get('sasl_config')
        if ssl_config and type(ssl_config) is dict:
            config['ssl_config'] = SSLConfig(**ssl_config)
        if sasl_config and type(sasl_config) is dict:
            config['sasl_config'] = SASLConfig(**sasl_config)
        return config


class KafkaSink(BaseSink):
    config_class = KafkaConfig

    def init_client(self):
        self._print('Start initializing producer.')
        # Initialize kafka producer
        kwargs = dict(
            bootstrap_servers=self.config.bootstrap_server,
            api_version=self.config.api_version,
        )
        if self.config.security_protocol == SecurityProtocol.SSL:
            kwargs['security_protocol'] = SecurityProtocol.SSL
            kwargs['ssl_cafile'] = self.config.ssl_config.cafile
            kwargs['ssl_certfile'] = self.config.ssl_config.certfile
            kwargs['ssl_keyfile'] = self.config.ssl_config.keyfile
            kwargs['ssl_password'] = self.config.ssl_config.password
            kwargs['ssl_check_hostname'] = self.config.ssl_config.check_hostname
        elif self.config.security_protocol == SecurityProtocol.SASL_SSL:
            kwargs['security_protocol'] = SecurityProtocol.SASL_SSL
            kwargs['sasl_mechanism'] = self.config.sasl_config.mechanism
            kwargs['sasl_plain_username'] = self.config.sasl_config.username
            kwargs['sasl_plain_password'] = self.config.sasl_config.password

        self.producer = KafkaProducer(
            **kwargs
        )
        self._print('Finish initializing producer.')

    def write(self, data: Dict):
        self._print(f'Ingest data {data}, time={time.time()}')
        self.producer.send(
            self.config.topic,
            json.dumps(data).encode('utf-8'),
        )

    def batch_write(self, data: List[Dict]):
        if not data:
            return
        self._print(f'Batch ingest {len(data)} records, time={time.time()}. Sample: {data[0]}')
        for record in data:
            self.producer.send(
                self.config.topic,
                json.dumps(record).encode('utf-8'),
            )



from mage_ai.io.base import ExportWritePolicy, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.io.sql import BaseSQL
from pandas import DataFrame, Series
from pydruid.db.api import Connection, Cursor as CursorParent
from typing import Dict, List, Union, Any, IO

WRITE_NOT_SUPPORTED_EXCEPTION = Exception('write operations are not supported.')


class Cursor(CursorParent):
    # added because the pydruid.db.api.Cursor doesn't implement __enter__ and __exit__
    # and we use the cursor in a `with` in sql.py
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass


class ConnectionWrapper(Connection):
    def cursor(self):
        return Cursor(url=self.url, user=self.user, password=self.password)


class Druid(BaseSQL):
    """
    Handles data transfer between a Druid data warehouse and the Mage app.
    """

    def __init__(
            self,
            host: str,
            port: int,
            user: Union[str, None] = None,
            password: Union[str, None] = None,
            path: str = '/druid/v2/sql/',
            scheme: str = 'http',
            verbose=True,
            **kwargs,
    ) -> None:
        """
        Initializes the data loader.

        Args:
            host (str): The host of the druid broker server to connect to.
            port (int): Port on which the druid broker server is running.
            user (str): The user with which to connect to the server with.
            password (str): The login password for the user.
            path (str): Path to druid sql api.
            scheme (str): The scheme identifies the protocol to be used http or https.
            **kwargs: Additional settings for creating SQLAlchemy engine and connection
        """
        self._ctx = None
        super().__init__(
            verbose=verbose,
            host=host,
            port=port,
            user=user,
            password=password,
            path=path,
            scheme=scheme,
            **kwargs,
        )

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'Druid':
        return cls(
            host=config[ConfigKey.DRUID_HOST],
            port=config[ConfigKey.DRUID_PORT],
            user=config[ConfigKey.DRUID_USER],
            password=config[ConfigKey.DRUID_PASSWORD],
            path=config[ConfigKey.DRUID_PATH],
            scheme=config[ConfigKey.DRUID_SCHEME],
        )

    def open(self) -> None:
        with self.printer.print_msg('Opening connection to Druid warehouse'):
            connect_kwargs = dict(
                host=self.settings['host'],
                port=self.settings['port'],
                user=self.settings['user'],
                password=self.settings['password'],
                path=self.settings['path'],
                scheme=self.settings['scheme'],
            )
            print(connect_kwargs)
            self._ctx = ConnectionWrapper(**connect_kwargs)

    def table_exists(self, schema_name: str, table_name: str) -> bool:
        # Not used
        with self.conn.cursor() as cur:
            cur.execute(
                f'SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = \'{table_name}\''
            )
        return bool(cur.rowcount)

    def execute(self, query_string: str, **query_vars) -> None:
        super().execute(query_string, **query_vars)

    def execute_queries(self,
                        queries: List[str],
                        query_variables: List[Dict] = None,
                        commit: bool = False,
                        fetch_query_at_indexes: List[bool] = None) -> List:
        return super().execute_queries(queries, query_variables, commit, fetch_query_at_indexes)

    def fetch_query(self, cursor, query: str) -> Any:
        return super().fetch_query(cursor, query)

    def load(self,
             query_string: str,
             limit: int = QUERY_ROW_LIMIT,
             display_query: Union[str, None] = None,
             verbose: bool = True,
             **kwargs) -> DataFrame:
        return super().load(query_string, limit, display_query, verbose, **kwargs)

    def clean(self, column: Series, dtype: str) -> Series:
        return super().clean(column, dtype)

    def close(self) -> None:
        super().close()

    def commit(self) -> None:
        super().commit()

    @property
    def conn(self) -> Any:
        return super().conn

    def rollback(self) -> None:
        super().rollback()

    def upload_dataframe(self,
                         cursor,
                         df: DataFrame,
                         db_dtypes: List[str],
                         dtypes: List[str],
                         full_table_name: str,
                         buffer: Union[IO, None] = None) -> None:
        raise WRITE_NOT_SUPPORTED_EXCEPTION

    def export(self,
               df: DataFrame,
               schema_name: str,
               table_name: str,
               if_exists: ExportWritePolicy = ExportWritePolicy.REPLACE,
               index: bool = False, verbose: bool = True,
               query_string: Union[str, None] = None,
               drop_table_on_replace: bool = False,
               cascade_on_drop: bool = False,
               allow_reserved_words: bool = False,
               unique_conflict_method: str = None,
               unique_constraints: List[str] = None) -> None:
        raise WRITE_NOT_SUPPORTED_EXCEPTION



from mage_ai.api.presenters.BasePresenter import BasePresenter


class ProjectPresenter(BasePresenter):
    default_attributes = [
        'help_improve_mage',
        'latest_version',
        'name',
        'project_uuid',
        'version',
    ]



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_cleaner.transformer_actions.constants import (
    ActionType,
    Axis,
)
import pandas as pd
import re


class ReformatValuesSubRule:
    """
    Assumptions of ReformatValuesSubRule
    1. df will not contain any empty strings - all empty strings are converted to null types.
    This is handled in ImputeValues.
    2. column_types will contain the correct type value
    3. Every column in df is of dtype object and the entries must be used to infer type.
    This is not always the case, but this assumption simplifies code
    """

    def __init__(
        self,
        action_builder,
        clean_column_cache,
        column_types,
        df,
        exact_dtypes,
        statistics,
    ):
        self.df = df
        self.column_types = column_types
        self.exact_dtypes = exact_dtypes
        self.statistics = statistics
        self.action_builder = action_builder
        self.clean_column_cache = clean_column_cache
        self.matches = []

    def clean_column(self, column):
        """
        Removes all null entries from a specific column
        """
        return self.clean_column_cache.setdefault(column, self.df[column].dropna(axis=0))

    def evaluate(self, column):
        raise NotImplementedError('Children of ReformatValuesSubRule must override this method.')

    def get_suggestions(self):
        raise NotImplementedError('Children of ReformatValuesSubRule must override this method.')


# TODO: how do we detect pattern like capitalizing the first letter in the sentence or word?
class StandardizeCapitalizationSubRule(ReformatValuesSubRule):
    UPPERCASE_PATTERN = r'^[^a-z]*$'
    LOWERCASE_PATTERN = r'^[^A-Z]*$'
    NON_ALPH_PATTERN = r'[^A-Za-z]'
    ALPHABETICAL_TYPES = frozenset(
        (
            ColumnType.CATEGORY_HIGH_CARDINALITY,
            ColumnType.CATEGORY,
            ColumnType.TEXT,
            ColumnType.EMAIL,
        )
    )
    NON_ALPH_UB = 0.4
    ALPH_RATIO_LB = 0.6

    def __init__(
        self,
        action_builder,
        clean_column_cache,
        column_types,
        df,
        exact_dtypes,
        statistics,
    ):
        super().__init__(
            action_builder,
            clean_column_cache,
            column_types,
            df,
            exact_dtypes,
            statistics,
        )
        self.uppercase = []
        self.lowercase = []

    def filter_column_regex(self, df_column, regex_pattern):
        if df_column.empty:
            return 0, df_column
        meets_regex = df_column.str.match(regex_pattern)
        try:
            count = meets_regex.value_counts()[True]
        except KeyError:
            count = 0
        return count, df_column[~meets_regex]

    def evaluate(self, column):
        """
        Rule:
        1. If column is not a category/string type which may have alphabet, no suggestion
        2. If non-null entries are not string, no suggestion
        3. If NON_ALPH_UB of entries are not majority alphabetical, no suggestion. Majority
            alphabetical == ALPH_RATIO_LB of all chars are alphabet
        4. If all entries are same case, no suggestion
        5. Suggest the more prevalent occurrence (e.g., if most alphabetical entries are lowercase
        text but some mixedcase and uppercase text, suggest conversion to lowercase)
        5a. If most alphabetical entries are mixedcase, suggest conversion to lowercase
        """
        dtype = self.column_types[column]
        if dtype not in self.ALPHABETICAL_TYPES:
            return

        clean_col = self.clean_column(column)
        if self.exact_dtypes[column] is not str:
            return

        non_alpha_ratio = clean_col.str.count(self.NON_ALPH_PATTERN) / clean_col.str.len()
        unfiltered_length = self.statistics[f'{column}/count']
        clean_col = clean_col[non_alpha_ratio <= self.NON_ALPH_UB]
        new_length = clean_col.count()
        if new_length / unfiltered_length <= self.ALPH_RATIO_LB:
            return

        uppercase, clean_col = self.filter_column_regex(clean_col, self.UPPERCASE_PATTERN)
        lowercase, clean_col = self.filter_column_regex(clean_col, self.LOWERCASE_PATTERN)
        mixedcase = clean_col.count()

        uppercase_ratio = uppercase / new_length
        lowercase_ratio = lowercase / new_length
        mixedcase_ratio = mixedcase / new_length

        if uppercase_ratio != 1 and lowercase_ratio != 1:
            max_case_style = max(uppercase_ratio, lowercase_ratio, mixedcase_ratio)
            if max_case_style == uppercase_ratio:
                self.uppercase.append(column)
            else:
                # if mixed case, default to lowercase
                self.lowercase.append(column)

    def get_suggestions(self):
        suggestions = []
        payloads = {'uppercase': self.uppercase, 'lowercase': self.lowercase}
        for case in payloads:
            if len(payloads[case]) != 0:
                suggestions.append(
                    self.action_builder(
                        'Reformat values',
                        f'Format entries in these columns as fully {case} to improve data quality.',
                        ActionType.REFORMAT,
                        action_arguments=payloads[case],
                        axis=Axis.COLUMN,
                        action_options={
                            'reformat': 'caps_standardization',
                            'capitalization': case,
                        },
                    )
                )
        return suggestions


class ConvertCurrencySubRule(ReformatValuesSubRule):
    CURR_PREFIX = r'(?:[\$\€\¥\₹\£]|(?:Rs)|(?:CAD))'
    CURR_SUFFIX = r'(?:[\元\€\$]|(?:CAD))'
    NUMBER_PATTERN = r'[0-9]*\.{0,1}[0-9]+'
    CURRENCY_BODY = rf'(?:{CURR_PREFIX}\s*{NUMBER_PATTERN}|{NUMBER_PATTERN}\s*{CURR_SUFFIX})'
    CURRENCY_PATTERN = re.compile(rf'^\s*(?:\-*\s*{CURRENCY_BODY}|{CURRENCY_BODY}\s*)\s*$')
    CURRENCY_TYPES = frozenset(
        (
            ColumnType.CATEGORY,
            ColumnType.CATEGORY_HIGH_CARDINALITY,
            ColumnType.TEXT,
            ColumnType.NUMBER,
            ColumnType.NUMBER_WITH_DECIMALS,
        )
    )

    def evaluate(self, column):
        """
        Rule:
        1. If the column is not a text, number, or category type, no suggestion
        2. If the column is not a string type, it can't contain currency symbol, no suggestion
        3. If all entries are of currency type (currency symbol followed by number),
           suggest conversion to number_with_decimal; else don't
        """
        dtype = self.column_types[column]
        if dtype not in self.CURRENCY_TYPES:
            return
        clean_col = self.clean_column(column)
        if self.exact_dtypes[column] is not str:
            return
        currency_pattern_mask = clean_col.str.match(self.CURRENCY_PATTERN)
        try:
            count = currency_pattern_mask.value_counts()[True]
        except KeyError:
            count = 0
        if count / self.statistics[f'{column}/count'] == 1:
            self.matches.append(column)

    def get_suggestions(self):
        suggestions = []
        if len(self.matches) != 0:
            suggestions.append(
                self.action_builder(
                    'Reformat values',
                    'Format entries in these columns as numbers to improve data quality.',
                    ActionType.REFORMAT,
                    action_arguments=self.matches,
                    axis=Axis.COLUMN,
                    action_options={
                        'reformat': 'currency_to_num',
                    },
                )
            )
        return suggestions


class ReformatDateSubRule(ReformatValuesSubRule):
    DATE_MATCHES_LB = 0.3
    DATE_TYPES = frozenset((ColumnType.DATETIME,))

    def evaluate(self, column):
        """
        Rule:
        1. If column is not of dtype category or text, no suggestion
        2. If column does not contain string types, no suggestion
        3. Try use Pandas datetime parse to convert from string to datetime.
           If more than DATE_MATCHES_LB entries are succesfully converted, suggest
           conversion to datetime type
        """
        dtype = self.column_types[column]
        if dtype not in self.DATE_TYPES:
            return
        if not self.exact_dtypes[column] is str:
            return
        clean_col = self.strip_column_for_date_parsing(column)
        clean_col = pd.to_datetime(clean_col, infer_datetime_format=True, errors='coerce')
        notnull_value_rate = clean_col.count() / len(clean_col)
        if notnull_value_rate >= self.DATE_MATCHES_LB:
            self.matches.append(column)

    def get_suggestions(self):
        suggestions = []
        if len(self.matches) != 0:
            suggestions.append(
                self.action_builder(
                    'Reformat values',
                    'Format entries in these columns as datetime objects to improve data quality.',
                    ActionType.REFORMAT,
                    action_arguments=self.matches,
                    axis=Axis.COLUMN,
                    action_options={
                        'reformat': 'date_format_conversion',
                    },
                )
            )
        return suggestions

    def strip_column_for_date_parsing(self, column):
        clean_col = self.clean_column(column)
        clean_col = clean_col.str.replace(r'[\,\s\t]+', ' ')
        clean_col = clean_col.str.replace(r'\s*([\/\\\-\.]+)\s*', lambda group: group.group(1)[0])
        return clean_col.str.lower()


class ReformatValues(BaseRule):
    RULE_LIST = [
        StandardizeCapitalizationSubRule,
        ConvertCurrencySubRule,
        ReformatDateSubRule,
    ]

    def __init__(self, df, column_types, statistics, custom_config={}):
        super().__init__(df, column_types, statistics, custom_config=custom_config)
        # TODO Clean dataframe prior to giving to rule
        self.clean_column_cache = {}
        self.exact_dtypes = self.infer_exact_dtypes()

    def hydrate_rule_list(self):
        return list(
            map(
                lambda x: x(
                    self._build_transformer_action_suggestion,
                    self.clean_column_cache,
                    self.column_types,
                    self.df,
                    self.exact_dtypes,
                    self.statistics,
                ),
                self.RULE_LIST,
            )
        )

    def evaluate(self):
        rules = self.hydrate_rule_list()
        for column in self.df_columns:
            for rule in rules:
                rule.evaluate(column)
        suggestions = []
        for rule in rules:
            suggestions.extend(rule.get_suggestions())
        return suggestions

    def infer_exact_dtypes(self):
        exact_dtypes = {}
        for column in self.df_columns:
            clean_col = self.df[column].dropna(axis=0)
            try:
                dtype = type(clean_col.iloc[0])
            except IndexError:
                dtype = None
            exact_dtypes[column] = dtype
        return exact_dtypes



from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.block.dbt.utils import (
    build_command_line_arguments,
    create_upstream_tables,
    create_temporary_profile,
    fetch_model_data,
    load_profiles_async,
    parse_attributes,
    query_from_compiled_sql,
    run_dbt_tests,
    update_model_settings,
)
from mage_ai.data_preparation.models.constants import BlockLanguage, BlockType
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.shared.hash import merge_dict
from typing import Any, Dict, List
import json
import os
import shutil
import subprocess


class DBTBlock(Block):
    @property
    def file_path(self) -> str:
        if BlockLanguage.SQL == self.language:
            repo_path = self.pipeline.repo_path if self.pipeline is not None else get_repo_path()
            file_path = self.configuration.get('file_path')

            return os.path.join(
                repo_path or os.getcwd(),
                'dbt',
                file_path,
            )

        return super().file_path

    async def metadata_async(self) -> Dict:
        project = None
        projects = {}

        if self.configuration.get('file_path'):
            attributes_dict = parse_attributes(self)
            profiles_full_path = attributes_dict['profiles_full_path']
            project_name = attributes_dict['project_name']
            profile_name = attributes_dict['profile_name']
            project = project_name

            targets = []
            profiles = await load_profiles_async(profile_name, profiles_full_path)
            outputs = profiles.get('outputs')
            if outputs:
                targets += sorted(list(outputs.keys()))

            projects[project_name] = dict(
                target=profiles.get('target'),
                targets=targets,
            )
        else:
            dbt_dir = os.path.join(get_repo_path(), 'dbt')
            project_names = [
                name for name in os.listdir(dbt_dir) if os.path.isdir(os.path.join(dbt_dir, name))
            ]
            for project_name in project_names:
                profiles_full_path = os.path.join(dbt_dir, project_name, 'profiles.yml')
                targets = []
                profiles = await load_profiles_async(project_name, profiles_full_path)
                outputs = profiles.get('outputs')
                if outputs:
                    targets += sorted(list(outputs.keys()))

                projects[project_name] = dict(
                    target=profiles.get('target'),
                    targets=targets,
                )

        return dict(dbt=dict(
            project=project,
            projects=projects,
        ))

    def run_tests(
        self,
        build_block_output_stdout=None,
        global_vars=None,
        **kwargs,
    ):
        run_dbt_tests(
            block=self,
            build_block_output_stdout=build_block_output_stdout,
            global_vars=global_vars,
        )

    def update_upstream_blocks(self, upstream_blocks: List[Any]) -> None:
        upstream_blocks_previous = self.upstream_blocks
        super().update_upstream_blocks(upstream_blocks)
        if BlockLanguage.SQL == self.language:
            update_model_settings(self, upstream_blocks, upstream_blocks_previous)

    def _execute_block(
        self,
        outputs_from_input_vars,
        execution_partition: str = None,
        global_vars: Dict = None,
        test_execution: bool = False,
        runtime_arguments: Dict = None,
        run_settings: Dict = None,
        **kwargs,
    ) -> List:
        variables = merge_dict(global_vars, runtime_arguments or {})

        if run_settings:
            test_execution = not (
                run_settings.get('build_model', False) or
                run_settings.get('run_model', False) or
                run_settings.get('test_model', False)
            )

        dbt_command, args, command_line_dict = build_command_line_arguments(
            self,
            variables,
            run_settings=run_settings,
            test_execution=test_execution,
        )

        project_full_path = command_line_dict['project_full_path']
        profiles_dir = command_line_dict['profiles_dir']
        dbt_profile_target = command_line_dict['profile_target']

        # Create a temporary profiles file with variables and secrets interpolated
        _, temp_profile_full_path = create_temporary_profile(
            project_full_path,
            profiles_dir,
        )

        is_sql = BlockLanguage.SQL == self.language
        if is_sql:
            create_upstream_tables(
                self,
                execution_partition=execution_partition,
                profile_target=dbt_profile_target,
                # TODO (tommy dang): this is creating unnecessary tables in notebook
                # cache_upstream_dbt_models=test_execution,
            )

        stdout = None if test_execution else subprocess.PIPE

        cmds = [
            'dbt',
            dbt_command,
        ] + args

        outputs = []

        args_pairs = []
        args_pair = []
        for a in args:
            args_pair.append(f"'{a}'" if '"' in a else a)
            if len(args_pair) == 2:
                args_pairs.append(args_pair)
                args_pair = []
        if len(args_pair) >= 1:
            args_pairs.append(args_pair)

        args_pairs = [f"  {' '.join(p)}" for p in args_pairs]
        args_string = ' \\\n'.join(args_pairs)

        print(f'dbt {dbt_command} \\\n{args_string}\n')

        if is_sql and test_execution:
            subprocess.run(
                cmds,
                preexec_fn=os.setsid,  # os.setsid doesn't work on Windows
                stdout=stdout,
            )
            df = query_from_compiled_sql(
                self,
                dbt_profile_target,
                limit=self.configuration.get('limit'),
            )
            self.store_variables(
                dict(df=df),
                execution_partition=execution_partition,
                override_outputs=True,
            )
            outputs = [df]
        elif not test_execution:
            proc = subprocess.Popen(
                cmds,
                bufsize=1,
                preexec_fn=os.setsid,  # os.setsid doesn't work on Windows
                stdout=stdout,
                universal_newlines=True,
            )
            for line in proc.stdout:
                print(line, end='')
            proc.communicate()
            if proc.returncode != 0 and proc.returncode is not None:
                raise subprocess.CalledProcessError(proc.returncode, proc.args)

        if not test_execution:
            run_results_file_path = os.path.join(project_full_path, 'target', 'run_results.json')
            with open(run_results_file_path, 'r') as f:
                try:
                    run_results = json.load(f)

                    print(f'\n{json.dumps(run_results, indent=2)}\n')

                    for result in run_results['results']:
                        if 'error' == result['status']:
                            raise Exception(result['message'])
                except json.decoder.JSONDecodeError:
                    print(f'WARNING: no run results found at {run_results_file_path}.')

            if is_sql and dbt_command in ['build', 'run']:
                limit = 1000
                if self.downstream_blocks and \
                        len(self.downstream_blocks) >= 1 and \
                        not all([BlockType.DBT == block.type for block in self.downstream_blocks]):
                    limit = None

                try:
                    df = fetch_model_data(
                        self,
                        dbt_profile_target,
                        limit=limit,
                    )

                    self.store_variables(
                        dict(output_0=df),
                        execution_partition=execution_partition,
                        override_outputs=True,
                    )
                    outputs = [df]
                except Exception as err:
                    print(f'Error: {err}')

        try:
            shutil.rmtree(profiles_dir)
        except Exception as err:
            print(f'Error removing temporary profile at {temp_profile_full_path}: {err}')

        return outputs



from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.triggers import ScheduleType
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.schedules import PipelineRun, PipelineSchedule
from mage_ai.orchestration.triggers.api import create_and_start_pipeline_run
from mage_ai.server.api.base import BaseHandler
from mage_ai.server.api.errors import UnauthenticatedRequestException
import json


class ApiTriggerPipelineHandler(BaseHandler):
    model_class = PipelineRun

    @safe_db_query
    def post(self, pipeline_schedule_id, token: str = None):
        pipeline_schedule = PipelineSchedule.query.get(int(pipeline_schedule_id))

        if ScheduleType.API == pipeline_schedule.schedule_type and \
            pipeline_schedule.token and \
                pipeline_schedule.token != token:
            raise UnauthenticatedRequestException(
                f'Invalid token for pipeline schedule ID {pipeline_schedule_id}.',
            )

        payload = self.get_payload()
        if 'variables' not in payload:
            payload['variables'] = {}

        body = self.request.body
        if body:
            payload['event_variables'] = {}

            for k, v in json.loads(body).items():
                if k == 'pipeline_run':
                    continue
                payload['event_variables'][k] = v

        pipeline = Pipeline.get(pipeline_schedule.pipeline_uuid)
        pipeline_run = create_and_start_pipeline_run(
            pipeline,
            pipeline_schedule,
            payload,
        )

        self.write(dict(pipeline_run=pipeline_run.to_dict()))



from datetime import datetime, timedelta
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.triggers import ScheduleInterval
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db.models.schedules import BlockRun, PipelineRun, PipelineSchedule
from mage_ai.shared.array import find

PIPELINE_FAILURE_STATUSES = [
    PipelineRun.PipelineRunStatus.CANCELLED,
    PipelineRun.PipelineRunStatus.FAILED,
]

BLOCK_FAILURE_STATUSES = [
    BlockRun.BlockRunStatus.CANCELLED,
    BlockRun.BlockRunStatus.FAILED,
]


def check_status(
    pipeline_uuid: str,
    execution_date: datetime,
    block_uuid: str = None,
) -> bool:
    __validate_pipeline_and_block(pipeline_uuid, block_uuid)

    pipeline_run = (
        PipelineRun
        .query
        .join(PipelineRun.pipeline_schedule)
        .filter(PipelineSchedule.pipeline_uuid == pipeline_uuid)
        .filter(PipelineRun.execution_date >= execution_date - timedelta(days=1))
        .order_by(PipelineRun.execution_date.desc())
        .first()
    )

    if pipeline_run is None:
        return False

    if pipeline_run.execution_date != execution_date and \
        pipeline_run.pipeline_schedule.schedule_interval != \
            ScheduleInterval.ONCE:
        return False

    pipeline_run.refresh()

    if block_uuid is not None:
        block_run = find(lambda run: run.block_uuid == block_uuid, pipeline_run.block_runs)
        if block_run is None:
            return False

        block_run.refresh()
        if block_run.status in BLOCK_FAILURE_STATUSES:
            raise Exception('Upstream block run failed, stopping sensor...')
        else:
            return block_run.status == BlockRun.BlockRunStatus.COMPLETED
    else:
        if pipeline_run.status in PIPELINE_FAILURE_STATUSES:
            raise Exception('Upstream pipeline run failed, stopping sensor...')
        else:
            return pipeline_run.status == PipelineRun.PipelineRunStatus.COMPLETED


def __validate_pipeline_and_block(pipeline_uuid, block_uuid):
    if pipeline_uuid not in Pipeline.get_all_pipelines(get_repo_path()):
        raise Exception('Pipeline not found, stopping sensor...')

    pipeline = Pipeline(pipeline_uuid, get_repo_path())

    if block_uuid is not None:
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception('Block not found in pipeline, stopping sensor...')



{% extends "data_loaders/default.jinja" %}
{% block imports %}
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.s3 import S3
from os import path
{{ super() -}}
{% endblock %}


{% block content %}
@data_loader
def load_from_s3_bucket(*args, **kwargs):
    """
    Template for loading data from a S3 bucket.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://docs.mage.ai/design/data-loading#s3
    """
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    bucket_name = 'your_bucket_name'
    object_key = 'your_object_key'

    return S3.with_config(ConfigFileLoader(config_path, config_profile)).load(
        bucket_name,
        object_key,
    )
{% endblock %}



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.VariablePresenter import VariablePresenter
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db.models.oauth import Permission


class VariablePolicy(BasePolicy):
    @property
    def entity(self):
        parent_model = self.options.get('parent_model')
        if parent_model:
            return Permission.Entity.PIPELINE, parent_model.uuid

        return Permission.Entity.PROJECT, get_repo_path()


VariablePolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

VariablePolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_notebook_edit_access())

VariablePolicy.allow_read(VariablePresenter.default_attributes + [
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

VariablePolicy.allow_read(VariablePresenter.default_attributes + [
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_notebook_edit_access())

VariablePolicy.allow_write([
    'name',
    'value',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_notebook_edit_access())



import io
import pandas as pd
import requests
from pandas import DataFrame

if 'data_loader' not in globals():
    from mage_ai.data_preparation.decorators import data_loader
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@data_loader
def load_data_from_api(**kwargs) -> DataFrame:
    """
    Template for loading data from API
    """
    url = 'https://raw.githubusercontent.com/datasciencedojo/datasets/master/titanic.csv'

    response = requests.get(url)
    return pd.read_csv(io.StringIO(response.text), sep=',')


@test
def test_output(df) -> None:
    """
    Template code for testing the output of the block.
    """
    assert df is not None, 'The output is undefined'



"""Add preferences to user

Revision ID: 1a737bdee53b
Revises: 847ea6602d6f
Create Date: 2023-04-03 11:03:27.102494

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '1a737bdee53b'
down_revision = '847ea6602d6f'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('user', sa.Column('preferences', sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('user', 'preferences')
    # ### end Alembic commands ###



from .base import BaseHandler
from enum import Enum
from mage_ai.cluster_manager.constants import (
    ECS_CLUSTER_NAME,
    ECS_CONTAINER_NAME,
    ECS_TASK_DEFINITION,
    GCP_PATH_TO_KEYFILE,
    GCP_PROJECT_ID,
    GCP_REGION,
    KUBE_NAMESPACE,
    KUBE_STORAGE_CLASS_NAME
)
from mage_ai.server.logger import Logger
import os
import yaml

logger = Logger().new_server_logger(__name__)


class ClusterType(str, Enum):
    EMR = 'emr'
    ECS = 'ecs'
    CLOUD_RUN = 'cloud_run'
    K8S = 'k8s'


class ApiInstancesHandler(BaseHandler):
    def get(self, cluster_type):
        instances = []
        if cluster_type == ClusterType.ECS:
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager

            try:
                cluster_name = self.get_argument('cluster_name', os.getenv(ECS_CLUSTER_NAME))
                ecs_instance_manager = EcsTaskManager(cluster_name)
                instances = ecs_instance_manager.list_tasks()
            except Exception as e:
                logger.error(str(e))
                instances = list()
        elif cluster_type == ClusterType.CLOUD_RUN:
            from mage_ai.cluster_manager.gcp.cloud_run_service_manager import CloudRunServiceManager
            project_id = os.getenv(GCP_PROJECT_ID)
            path_to_credentials = os.getenv(GCP_PATH_TO_KEYFILE)
            region = os.getenv(GCP_REGION)
            cloud_run_service_manager = CloudRunServiceManager(
                project_id,
                path_to_credentials,
                region=region
            )

            instances = cloud_run_service_manager.list_services()
        elif cluster_type == ClusterType.K8S:
            from mage_ai.cluster_manager.kubernetes.workload_manager import WorkloadManager
            namespace = os.getenv(KUBE_NAMESPACE)
            workload_manager = WorkloadManager(namespace)

            instances = workload_manager.list_services()

        self.write(dict(instances=instances))

    def post(self, cluster_type):
        try:
            if cluster_type == ClusterType.ECS:
                from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager
                instance_payload = self.get_payload().get('instance')
                name = instance_payload.get('name')
                cluster_name = instance_payload.get('cluster_name', os.getenv(ECS_CLUSTER_NAME))
                task_definition = instance_payload.get(
                    'task_definition',
                    os.getenv(ECS_TASK_DEFINITION),
                )
                container_name = instance_payload.get(
                    'container_name',
                    os.getenv(ECS_CONTAINER_NAME),
                )

                ecs_instance_manager = EcsTaskManager(cluster_name)

                instance = ecs_instance_manager.create_task(
                    name,
                    task_definition,
                    container_name,
                )

                self.write(dict(
                    instance=instance,
                    success=True,
                ))
            elif cluster_type == ClusterType.CLOUD_RUN:
                from mage_ai.cluster_manager.gcp.cloud_run_service_manager \
                    import CloudRunServiceManager
                instance_payload = self.get_payload().get('instance')
                name = instance_payload.get('name')
                project_id = instance_payload.get('project_id', os.getenv(GCP_PROJECT_ID))
                path_to_credentials = instance_payload.get(
                    'path_to_credentials',
                    os.getenv('path_to_keyfile')
                )
                region = instance_payload.get('region', os.getenv('GCP_REGION'))

                cloud_run_service_manager = CloudRunServiceManager(
                    project_id,
                    path_to_credentials,
                    region=region
                )

                cloud_run_service_manager.create_service(name)

                self.write(dict(success=True))
            elif cluster_type == ClusterType.K8S:
                from mage_ai.cluster_manager.kubernetes.workload_manager import WorkloadManager
                instance_payload = self.get_payload().get('instance')
                name = instance_payload.get('name')
                namespace = instance_payload.get('namespace', os.getenv(KUBE_NAMESPACE))
                storage_class_name = instance_payload.get(
                    'storage_class_name',
                    os.getenv(KUBE_STORAGE_CLASS_NAME)
                )
                service_account_name = instance_payload.get('service_account_name')
                container_config_yaml = instance_payload.get('container_config')
                container_config = None
                if container_config_yaml:
                    container_config = yaml.full_load(container_config_yaml)

                k8s_workload_manager = WorkloadManager(namespace)
                k8s_workload_manager.create_stateful_set(
                    name,
                    container_config=container_config,
                    service_account_name=service_account_name,
                    storage_class_name=storage_class_name,
                )
                self.write(dict(success=True))
        except Exception as e:
            self.write(dict(success=False, error_message=str(e)))


class ApiInstanceDetailHandler(BaseHandler):
    def put(self, cluster_type, instance_name):
        if cluster_type == 'ecs':
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager
            instance_payload = self.get_payload().get('instance')
            task_arn = instance_payload.get('task_arn')
            cluster_name = instance_payload.get('cluster_name', os.getenv(ECS_CLUSTER_NAME))
            task_definition = instance_payload.get(
                'task_definition',
                os.getenv(ECS_TASK_DEFINITION),
            )
            container_name = instance_payload.get('container_name', os.getenv(ECS_CONTAINER_NAME))

            action = instance_payload.get('action')

            ecs_instance_manager = EcsTaskManager(cluster_name)
            instance = None
            if action == 'stop':
                instance = ecs_instance_manager.stop_task(task_arn)
            elif action == 'resume':
                instance = ecs_instance_manager.create_task(
                    instance_name,
                    task_definition,
                    container_name,
                )

        self.write(dict(
            instance=instance,
            success=True,
        ))

    def delete(self, cluster_type, instance_name):
        if cluster_type == 'ecs':
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager
            task_arn = self.get_argument('task_arn', None)
            cluster_name = self.get_argument('cluster_name', os.getenv(ECS_CLUSTER_NAME))

            ecs_instance_manager = EcsTaskManager(cluster_name)
            ecs_instance_manager.delete_task(instance_name, task_arn)

        self.write(dict(
            success=True
        ))



from datetime import datetime
from mage_ai.server.active_kernel import get_active_kernel_client
from mage_ai.server.logger import Logger

logger = Logger().new_server_logger(__name__)


def get_messages(callback=None):
    now = datetime.utcnow()

    while True:
        try:
            client = get_active_kernel_client()
            message = client.get_iopub_msg(timeout=1)

            if message.get('content'):
                if callback:
                    callback(message)
                else:
                    logger.warn(f'[{now}] No callback for message: {message}')
        except Exception as e:
            if str(e):
                logger.error(f'[{now}] Error: {e}', )
            pass



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sinks.base import BaseSink
from typing import Dict, List
import boto3
import json
import time


@dataclass
class KinesisConfig(BaseConfig):
    stream_name: str        # Kinesis stream name
    partition_key: str      # The partition key to use for the data.


class KinesisSink(BaseSink):
    config_class = KinesisConfig

    def init_client(self):
        self.kinesis_client = boto3.client('kinesis')

    def test_connection(self):
        return True

    def write(self, data: Dict):
        self._print(f'Ingest data {data}, time={time.time()}')
        self.kinesis_client.put_record(
            StreamName=self.config.stream_name,
            Data=json.dumps(data),
            PartitionKey=self.config.partition_key,
        )

    def batch_write(self, data: List[Dict]):
        if not data:
            return
        self._print(f'Batch ingest {len(data)} records, time={time.time()}. Sample: {data[0]}')
        records = [
            {
                'Data': json.dumps(d).encode('utf-8'),
                'PartitionKey': self.config.partition_key,
            } for d in data
        ]
        self.kinesis_client.put_records(
            Records=records,
            StreamName=self.config.stream_name,
        )



from mage_ai.api.presenters.BasePresenter import BasePresenter


class ExtensionOptionPresenter(BasePresenter):
    default_attributes = [
        'description',
        'name',
        'templates',
        'uuid',
    ]



from mage_ai.orchestration.queue.config import QueueConfig
from mage_ai.orchestration.queue.queue import Queue
from typing import Callable


class CeleryQueue(Queue):
    """
    TODO: implement this class.
    """

    def __init__(self, queue_config: QueueConfig):
        # self.celery_app = Celery(
        #     'mypackage',
        #     broker='amqp://guest@localhost//',
        #     backend='amqp://guest@localhost//',
        # )
        pass

    def clean_up_jobs(self):
        pass

    def enqueue(self, job_id: str, target: Callable, *args, **kwargs):
        pass

    def has_job(self, job_id: str):
        return False

    def kill_job(self, job_id: str):
        pass



from faker import Faker
from mage_ai.data_cleaner.column_types.column_type_detector import (
    MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES,
    find_syntax_errors,
    infer_column_types,
    REGEX_NUMBER,
)
from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.shared.hash import merge_dict
from mage_ai.tests.base_test import TestCase
import pandas as pd
import numpy as np


class ColumnTypeDetectorTests(TestCase):
    def setUp(self):
        Faker.seed(42)
        self.fake = Faker()
        return super().setUp()

    def test_get_syntax_errors(self):
        df = pd.DataFrame(
            [
                [1, 'abc@xyz.com', '32132'],
                [2, 'abc2@xyz.com', '12345'],
                [3, 'test', '1234'],
                [4, 'abc@test.net', 'abcde'],
                [5, 'abc12345@', '54321'],
                [6, 'abcdef@123.com', '56789'],
            ],
            columns=['id', 'email', 'zip_code'],
        )
        rows1 = df['id'][find_syntax_errors(df['id'], 'number')].tolist()
        rows2 = df['email'][find_syntax_errors(df['email'], 'email')].tolist()
        rows3 = df['zip_code'][find_syntax_errors(df['zip_code'], 'zip_code')].tolist()
        self.assertEqual(rows1, [])
        self.assertEqual(rows2, ['test', 'abc12345@'])
        self.assertEqual(rows3, ['abcde'])

    def test_get_syntax_errors_invalid_null_values(self):
        df = pd.DataFrame(
            [
                ['1', 'abc@xyz.com', 'invalid', None, '12-22-34'],
                ['2', 'missing', None, 'invalid', '4-27-19 12:45:56'],
                [np.nan, 'test', '1234', None, 'not a date'],
                [0, 'invalid', 'abcde', 'phone number', pd.NaT],
                ['4', 'abc12345@', '54321', 'missing', pd.Timestamp.min],
                ['5', None, 'missing', 'another phone', None],
            ],
            columns=['id', 'email', 'zip_code', 'phone_number', 'dates'],
        )
        rows1 = df['id'][find_syntax_errors(df['id'], 'number')].tolist()
        rows2 = df['email'][find_syntax_errors(df['email'], 'email')].tolist()
        rows3 = df['zip_code'][find_syntax_errors(df['zip_code'], 'zip_code')].tolist()
        rows4 = df['phone_number'][find_syntax_errors(df['phone_number'], 'phone_number')].tolist()
        rows5 = df['dates'][find_syntax_errors(df['dates'], 'datetime')].tolist()
        self.assertEqual(rows1, [])
        self.assertEqual(rows2, ['test', 'abc12345@'])
        self.assertEqual(rows3, ['abcde'])
        self.assertEqual(rows4, ['phone number', 'another phone'])
        self.assertEqual(rows5, ['not a date'])

    def test_number_pattern_matching(self):
        good_number_patterns = [
            '3123452',
            '23.234232',
            '-23,245,234.93',
            '.934',
            '-0.000002',
            '$ 0.043',
            '-€ 23,456,789.00092',
            '¥ 21345634',
            '- Rs   345.23423523',
            'CAD 2,345,234,345.23423',
            '234.23423423 $',
            '-8907987 CAD',
            '3239458734 元',
            '234,2342,2343.823034234%',
            '234.3433   %',
        ]
        bad_number_patterns = [
            'this is not a number',
            'NOTCURR 23423933',
            '0.000.00.000.000',
            '1-111-111-111',
            ',2343234203',
            'ksjh&A(*2jnkj&#$#$',
        ]

        for good in good_number_patterns:
            match = REGEX_NUMBER.match(good)
            self.assertIsNotNone(match)

        for bad in bad_number_patterns:
            match = REGEX_NUMBER.match(bad)
            self.assertIsNone(match)

    def test_infer_column_types(self):
        df = self.__build_test_df()
        column_types = infer_column_types(df)

        self.assertEqual(
            column_types,
            {
                'true_or_false': ColumnType.TRUE_OR_FALSE,
                'number_with_decimals': ColumnType.NUMBER_WITH_DECIMALS,
                'category': ColumnType.CATEGORY,
                'datetime': ColumnType.DATETIME,
                'text': ColumnType.TEXT,
                'number': ColumnType.NUMBER,
                'number_with_dollars': ColumnType.NUMBER_WITH_DECIMALS,
                'number_with_percentage': ColumnType.NUMBER_WITH_DECIMALS,
                'zip_code': ColumnType.ZIP_CODE,
                'zip_code_with_3_numbers': ColumnType.ZIP_CODE,
                'invalid_zip_code': ColumnType.NUMBER,
                'email': ColumnType.EMAIL,
                'phone_number': ColumnType.PHONE_NUMBER,
                'datetime_abnormal': ColumnType.DATETIME,
                'name': ColumnType.TEXT,
            },
        )

    def test_infer_column_types_with_existing_column_types(self):
        df = self.__build_test_df()
        existing_column_types = {
            'true_or_false': ColumnType.TRUE_OR_FALSE,
            'number_with_decimals': ColumnType.NUMBER,
            'category': ColumnType.CATEGORY_HIGH_CARDINALITY,
            'datetime': ColumnType.DATETIME,
            'text': ColumnType.TEXT,
            'number': ColumnType.NUMBER_WITH_DECIMALS,
            'number_with_dollars': ColumnType.NUMBER,
            'number_with_percentage': ColumnType.NUMBER,
        }
        column_types = infer_column_types(df, column_types=existing_column_types)
        self.assertEqual(
            column_types,
            merge_dict(
                existing_column_types,
                {
                    'zip_code': ColumnType.ZIP_CODE,
                    'zip_code_with_3_numbers': ColumnType.ZIP_CODE,
                    'invalid_zip_code': ColumnType.NUMBER,
                    'email': ColumnType.EMAIL,
                    'phone_number': ColumnType.PHONE_NUMBER,
                    'datetime_abnormal': ColumnType.DATETIME,
                    'name': ColumnType.TEXT,
                },
            ),
        )

    def test_integer_recognition(self):
        max_int = np.iinfo(np.int64).max
        max_int_unsigned = np.iinfo(np.uint).max
        min_int = np.iinfo(np.int64).min
        df = pd.DataFrame(
            dict(
                integers=[1, 2, 3, 4, 5],
                not_integers=[1.2, 3.4, 5.6, 6.8, 9.2],
                more_integers=[1.0, 2.0, 3.0, 4.0, 5.0],
                integers_with_null=[1, None, 2, 3, np.nan],
                not_integers_with_null=[1.1, np.nan, 2.0, 3.0, np.nan],
                str_integers=['1.0', None, '2.0', '3.0', None],
                str_not_integers=['1.000', None, '2.0002', '3.000', None],
                large_integers=[max_int, min_int, max_int - 1, min_int + 1, max_int],
                large_integers_unsigned=[
                    max_int,
                    max_int_unsigned,
                    max_int - 1,
                    0,
                    max_int_unsigned,
                ],
                str_large_integers=[
                    str(max_int),
                    str(min_int),
                    str(max_int - 1),
                    str(min_int + 1),
                    str(max_int),
                ],
                too_large_integers=[
                    str(max_int),
                    str(min_int),
                    str(max_int + 1),
                    str(min_int - 1),
                    str(max_int),
                ],
                too_large_integers_version_2=[
                    str(max_int),
                    str(min_int),
                    str(max_int_unsigned),
                    str(min_int - 1),
                    str(max_int),
                ],
            )
        )
        ctypes = infer_column_types(df)
        expected_ctypes = dict(
            integers='number',
            not_integers='number_with_decimals',
            more_integers='number',
            integers_with_null='number',
            not_integers_with_null='number_with_decimals',
            str_integers='number',
            str_not_integers='number_with_decimals',
            large_integers='number',
            large_integers_unsigned='number',
            str_large_integers='number',
            too_large_integers='category',
            too_large_integers_version_2='category',
        )
        self.assertEqual(ctypes, expected_ctypes)

    def test_phone_number_recognition(self):
        df = pd.DataFrame(
            dict(
                phone_num=[self.fake.phone_number() for _ in range(6)],
                str_int_phone_nums=[
                    '9999999999',
                    '12345678909',
                    '918328328322',
                    '33459830234',
                    '8234592345',
                    '23439433343',
                ],
                not_phone_nums=[
                    '99999',
                    '12349',
                    '918328328322',
                    '33430234',
                    '8234592345',
                    '23439433343',
                ],
                int_phone_nums=[
                    9999999999,
                    12345678909,
                    918328328322,
                    33459830234,
                    8234592345,
                    23439433343,
                ],
                float_phone_nums=[
                    9999999999,
                    np.nan,
                    918328328322.0,
                    33459830234,
                    8234592345,
                    np.nan,
                ],
                not_float_phone_nums=[
                    9999999999.00234233,
                    np.nan,
                    918328328322.0,
                    33459830234.234,
                    8234592345,
                    np.nan,
                ],
                bad_column_name_one=[
                    9999999999,
                    np.nan,
                    918328328322.0,
                    33459830234,
                    8234592345,
                    np.nan,
                ],
                bad_column_name_two=[self.fake.phone_number() for _ in range(6)],
            )
        )
        ctypes = infer_column_types(df)
        expected_ctypes = dict(
            phone_num='phone_number',
            str_int_phone_nums='phone_number',
            not_phone_nums='number',
            int_phone_nums='phone_number',
            float_phone_nums='phone_number',
            not_float_phone_nums='number_with_decimals',
            bad_column_name_one='number',
            bad_column_name_two='text',
        )
        self.assertEqual(ctypes, expected_ctypes)

    def test_zip_code_recognition(self):
        df = pd.DataFrame(
            dict(
                zip_code=[self.fake.postcode() for _ in range(6)],
                postal_code=[self.fake.postcode() for _ in range(6)],
                bad_column_name=[self.fake.postcode() for _ in range(6)],
                string_zips=[
                    '12345',
                    '132',
                    '234-3434',
                    '765',
                    '41324-2343',
                    '12342',
                ],
                not_string_postal_zips=[
                    '12',
                    '13234523452',
                    '23-3434',
                    '765,23453',
                    '41324-2343-23423',
                    '12342',
                ],
                int_zips=[
                    12345,
                    53123,
                    323,
                    423,
                    678,
                    67896,
                ],
                not_int_zips=[
                    12345,
                    34553123,
                    323343,
                    423,
                    67834534,
                    67896,
                ],
                bad_column_name_two=[
                    12345,
                    53123,
                    323,
                    423,
                    678,
                    67896,
                ],
            )
        )
        ctypes = infer_column_types(df)
        expected_ctypes = dict(
            zip_code=ColumnType.ZIP_CODE,
            postal_code=ColumnType.ZIP_CODE,
            bad_column_name=ColumnType.NUMBER,
            string_zips=ColumnType.ZIP_CODE,
            not_string_postal_zips=ColumnType.TEXT,
            int_zips=ColumnType.ZIP_CODE,
            not_int_zips=ColumnType.NUMBER,
            bad_column_name_two=ColumnType.NUMBER,
        )
        self.assertEqual(ctypes, expected_ctypes)

    def test_datetime_recognition(self):
        df = pd.DataFrame(
            dict(
                dates=[self.fake.date() for _ in range(6)],
                datetimes=[self.fake.date_time().strftime('%d-%m-%Y %H:%M:%S') for _ in range(6)],
                datetimes_iso=[self.fake.date_time().isoformat() for _ in range(6)],
                dates_w_slash=[self.fake.date_time().strftime('%m/%d/%Y') for _ in range(6)],
                datetimes_w_slash=[
                    self.fake.date_time().strftime('%m/%d/%Y %H:%M:%S') for _ in range(6)
                ],
            )
        )
        ctypes = infer_column_types(df)
        for col in ctypes:
            self.assertEqual(ctypes[col], ColumnType.DATETIME)

    def __build_test_df(self):
        columns = [
            'true_or_false',
            'number_with_decimals',
            'category',
            'datetime',
            'text',
            'number',
            'number_with_dollars',
            'number_with_percentage',
            'zip_code',
            'zip_code_with_3_numbers',
            'invalid_zip_code',
            'email',
            'phone_number',
            'datetime_abnormal',
            'name',
        ]
        table = [
            [
                '1',
                '3',
                'male',
                '2020-1-1',
                '1.0',
                '1',
                3,
                '30%',
                '10128-1213',
                123,
                123,
                'fire@mage.ai',
                '123-456-7890',
                'May 4, 2021, 6:35 PM',
                self.fake.name(),
            ],
            [
                '1',
                '12.0',
                'female',
                '2020-07-13',
                ' '.join(['t' for i in range(MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES + 1)]),
                '2',
                '€4',
                '12.32%',
                12345,
                1234,
                1234,
                'mage@fire.com',
                '(123) 456-7890',
                'Feb 17, 2021, 2:57 PM',
                self.fake.name(),
            ],
            [
                '1',
                '0',
                'machine',
                '2020-06-25 01:02',
                ' '.join(['t' for i in range(MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES + 1)]),
                '3',
                '¥5,000',
                '50%',
                '12345',
                12345,
                12345,
                'fire_mage@mage.com',
                '1234567890',
                'Feb 18, 2021, 2:57 PM',
                self.fake.name(),
            ],
            [
                0,
                '40.7',
                'mutant',
                '2020-12-25 01:02:03',
                ' '.join(['t' for i in range(MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES + 1)]),
                '4',
                'Rs 5,000.01',
                '20%',
                '12345',
                12345,
                123456,
                'mage-fire@mage.ai',
                '1234567',
                'Feb 19, 2021, 2:57 PM',
                self.fake.name(),
            ],
            [
                0,
                '40.7',
                'alien',
                '2020-12-25T01:02:03.000Z',
                ' '.join(['t' for i in range(MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES + 1)]),
                '4',
                '-10128,121.3123元',
                '18%',
                '01234',
                12345,
                12,
                'fire+1@mage.ai',
                '(123)456-7890',
                'Feb 20, 2021, 2:57 PM',
                self.fake.name(),
            ],
        ]
        date_formats = [
            '01/1/2019',
            '1/1/2019',
            '1/21/2019',
            '11/1/2019',
            '2020/01/1',
            '2020/1/01',
            '2020/1/1',
            'Pending',
        ]

        for date_format in date_formats:
            table.append(
                [
                    0,
                    '40.7',
                    'mutant',
                    date_format,
                    self.fake.text(),
                    4,
                    '$5,000.01',
                    '15.32%',
                    '01234',
                    12345,
                    12,
                    'fire+1@mage.ai',
                    '(123)456-7890',
                    'Feb 18, 2021, 2:57 PM',
                    self.fake.name(),
                ]
            )

        df = pd.DataFrame(table, columns=columns)
        return df

    def test_list_recognition(self):
        df = pd.DataFrame(
            {
                'lists': [
                    ['this', 'is', 'a', 'list', 'of', 'strings'],
                    ['this', 'is', 'a', 'list', 'of', 'strings'],
                    ['this', 'is', 'a', 'list', 'of', 'strings'],
                    None,
                ],
                'lists2': [
                    [2, 1, 3, 4, 2, 1, 2, 2],
                    [8, 9, 6, 4, 6, 4, 5, 4, 3, 4],
                    [],
                    [2, 3, 4, 1, 2],
                ],
                'lists3': [
                    None,
                    [True, False, True, True],
                    [False, True, False, True],
                    [True, True, True, False],
                ],
                'lists4': [
                    [2, 'string', False, None],
                    [np.nan, 2.0, 'string', '3'],
                    ['not string?', True, True, 8, False, np.nan, None],
                    [],
                ],
                'tuples': [
                    (2, 'string', False, None),
                    (np.nan, 2.0, 'string', '3'),
                    ('not string?', True, True, 8, False, np.nan, None),
                    tuple(),
                ],
                'string_lists': [
                    '[2, \'string\', False, None]',
                    '[np.nan,2.0,\'string\',\'3\']',
                    '[\'not string?\'   ,  True, True , 8   , False  , np.nan, None]',
                    '[]',
                ],
                'string_tuples': [
                    '(2, \'string\', False, None)',
                    '(np.nan, 2.0, \'string\', \'3\')',
                    '(\'not string?\', True, True, 8, False, np.nan, None)',
                    '()',
                ],
                'not_a_list': [
                    '3',
                    4,
                    'a category',
                    'a very long piece of text',
                ],
            }
        )
        ctypes = infer_column_types(df)
        expected_ctypes = dict(
            lists=ColumnType.LIST,
            lists2=ColumnType.LIST,
            lists3=ColumnType.LIST,
            lists4=ColumnType.LIST,
            tuples=ColumnType.LIST,
            string_lists=ColumnType.LIST,
            string_tuples=ColumnType.LIST,
            not_a_list=ColumnType.TEXT,
        )
        self.assertEqual(ctypes, expected_ctypes)



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.RolePresenter import RolePresenter


class RolePolicy(BasePolicy):
    pass


RolePolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

RolePolicy.allow_read(RolePresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

RolePolicy.allow_query([
    'limit_roles',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())



from datetime import datetime
from unittest.mock import patch

from freezegun import freeze_time

from mage_ai.api.operations import constants
from mage_ai.authentication.ldap import LDAPConnection
from mage_ai.authentication.oauth2 import encode_token
from mage_ai.orchestration.db.models.oauth import Oauth2AccessToken, User
from mage_ai.tests.api.operations.test_base import BaseApiTestCase
from mage_ai.tests.factory import create_user


class SessionOperationTests(BaseApiTestCase):
    @freeze_time(datetime(3333, 12, 12))
    async def test_execute_create(self):
        password = 'password'
        user = create_user(password=password)

        operation = self.build_operation(
            action=constants.CREATE,
            payload=dict(session=dict(
                email=user.email,
                password=password,
            )),
            resource='sessions',
        )
        response = await operation.execute()

        access_token = Oauth2AccessToken.query.filter(Oauth2AccessToken.user_id == user.id).first()

        self.assertEqual(
            response['session']['token'],
            encode_token(access_token.token, access_token.expires),
        )

    async def test_execute_create_failed(self):
        password = 'password'
        user = create_user(password=password)

        operation = self.build_operation(
            action=constants.CREATE,
            payload=dict(session=dict(
                email=user.email,
                password='not password',
            )),
            resource='sessions',
        )
        response = await operation.execute()

        self.assertIsNotNone(response['error'])

    @patch('mage_ai.api.resources.SessionResource.AUTHENTICATION_MODE', 'ldap')
    @patch.object(LDAPConnection, 'authorize')
    @patch.object(LDAPConnection, 'authenticate')
    async def test_ldap_login(self, mock_authenticate, mock_authorize):
        mock_authenticate.return_value = (True, "Julius_Novachrono")
        mock_authorize.return_value = True

        username = "novachrono"
        operation = self.build_operation(
            action=constants.CREATE,
            payload=dict(session=dict(
                email=username,
                password="Wizard King",
            )),
            resource='sessions',
        )
        await operation.execute()

        mock_authenticate.assert_called_once_with("novachrono", "Wizard King")
        mock_authorize.assert_called_once_with("Julius_Novachrono")
        self.assertIsNotNone(User.query.filter(User.username == username).first())

    @patch('mage_ai.api.resources.SessionResource.AUTHENTICATION_MODE', 'ldap')
    @patch.object(LDAPConnection, 'authorize')
    @patch.object(LDAPConnection, 'authenticate')
    async def test_ldap_login_unauthenticated(self, mock_authenticate, mock_authorize):
        mock_authenticate.return_value = (False, "")
        mock_authorize.return_value = False

        username = "licht"
        operation = self.build_operation(
            action=constants.CREATE,
            payload=dict(session=dict(
                email=username,
                password="Golden Dawn",
            )),
            resource='sessions',
        )
        response = await operation.execute()

        mock_authenticate.assert_called_once_with("licht", "Golden Dawn")
        mock_authorize.assert_not_called()
        self.assertIsNone(User.query.filter(User.username == username).first())
        self.assertIsNotNone(response['error'])

    @patch('mage_ai.api.resources.SessionResource.AUTHENTICATION_MODE', 'ldap')
    @patch.object(LDAPConnection, 'authorize')
    @patch.object(LDAPConnection, 'authenticate')
    async def test_ldap_login_unauthorized(self, mock_authenticate, mock_authorize):
        mock_authenticate.return_value = (True, "Yami_Sukehiro")
        mock_authorize.return_value = False

        username = "yami"
        operation = self.build_operation(
            action=constants.CREATE,
            payload=dict(session=dict(
                email=username,
                password="black bull",
            )),
            resource='sessions',
        )
        response = await operation.execute()

        mock_authenticate.assert_called_once_with("yami", "black bull")
        mock_authorize.assert_called_once_with("Yami_Sukehiro")
        self.assertIsNone(User.query.filter(User.username == username).first())
        self.assertIsNotNone(response['error'])



from mage_ai.data_cleaner.cleaning_rules.remove_columns_with_single_value import (
    RemoveColumnsWithSingleValue,
)
from mage_ai.tests.base_test import TestCase
import pandas as pd
import numpy as np


class RemoveColumnWithSingleValueTests(TestCase):
    def test_evaluate(self):
        df = pd.DataFrame(
            [
                [1, '2022-01-01', True],
                [2, '2022-01-02', True],
                [3, np.NaN, True],
                [4, np.NaN, True],
                [5, np.NaN, True],
            ],
            columns=['id', 'deleted_at', 'is_active'],
        )
        column_types = {
            'id': 'number',
            'deleted_at': 'datetime',
            'is_active': 'true_or_false',
        }
        statistics = {
            'id/count_distinct': 5,
            'deleted_at/count_distinct': 2,
            'is_active/count_distinct': 1,
        }
        result = RemoveColumnsWithSingleValue(df, column_types, statistics).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Remove columns with single value',
                    message='Remove columns with a single unique value to reduce the amount of '
                    'redundant data.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='remove',
                        action_arguments=['is_active'],
                        action_code='',
                        action_options={},
                        action_variables={},
                        axis='column',
                        outputs=[],
                    ),
                )
            ],
        )



"""Add settings to pipeline schedule

Revision ID: 1d8e65aeefdd
Revises: 26305e46df52
Create Date: 2023-01-26 14:54:26.688979

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1d8e65aeefdd'
down_revision = '26305e46df52'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('pipeline_schedule', sa.Column('settings', sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('pipeline_schedule', 'settings')
    # ### end Alembic commands ###



from datetime import datetime
from mage_ai.shared.parsers import encode_complex
from mage_ai.shared.hash import merge_dict
from typing import Dict
import logging
import simplejson
import traceback
import uuid


class DictLogger():
    def __init__(self, logger: logging.Logger, logging_tags: Dict = dict()):
        self.logger = logger
        self.logging_tags = logging_tags

    def critical(self, message, **kwargs):
        self.__send_message('critical', message, **kwargs)

    def debug(self, message, **kwargs):
        self.__send_message('debug', message, **kwargs)

    def error(self, message, **kwargs):
        self.__send_message('error', message, **kwargs)

    def exception(self, message, **kwargs):
        self.__send_message('exception', message, **kwargs)

    def info(self, message, **kwargs):
        self.__send_message('info', message, **kwargs)

    def log(self, log_level, message, **kwargs):
        self.__send_message('log', message, log_level=log_level, **kwargs)

    def warning(self, message, **kwargs):
        self.__send_message('warning', message, **kwargs)

    def __send_message(
        self,
        method_name,
        message,
        error=None,
        log_level=None,
        **kwargs,
    ):
        now = datetime.utcnow()
        data = dict(
            level=method_name.upper(),
            message=message,
            timestamp=now.timestamp(),
            uuid=uuid.uuid4().hex,
        )

        if error:
            data['error'] = traceback.format_exc(),
            data['error_stack'] = traceback.format_stack(),
            data['error_stacktrace'] = str(error),

        msg = simplejson.dumps(
            merge_dict(self.logging_tags or dict(), merge_dict(kwargs, data)),
            default=encode_complex,
            ignore_nan=True,
        )

        if log_level is None:
            getattr(self.logger, method_name)(msg)
        else:
            getattr(self.logger, method_name)(log_level, msg)

    def __getattr__(self, method_name):
        if method_name not in [
            '__send_message',
            'critical',
            'debug',
            'error',
            'exception',
            'info',
            'log',
            'warning',
        ]:
            return getattr(self.logger, method_name)
        return getattr(self, method_name)



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.SHIFT_DOWN

    Shifts value in the selected column down by specified number periods.

    Docs: https://docs.mage.ai/guides/transformer-blocks#shift-down
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.SHIFT_DOWN,
        arguments=[],  # Specify one column to perform shift on
        axis=Axis.COLUMN,
        options={'periods': 1},
        outputs=[{'uuid': 'down_shifted_column_1', 'type': 'category'}],
    )
{% endblock %}



from mage_ai.orchestration.db.models.schedules import PipelineRun
import multiprocessing


class ExecutionProcessManager:
    def __init__(self):
        self.block_processes = dict()
        self.pipeline_processes = dict()

    def has_pipeline_process(self, pipeline_run_id: int):
        return (
            pipeline_run_id in self.pipeline_processes
            and self.pipeline_processes[pipeline_run_id].is_alive()
        )

    def terminate_pipeline_process(self, pipeline_run_id: int) -> None:
        if self.has_pipeline_process(pipeline_run_id):
            try:
                self.pipeline_processes[pipeline_run_id].terminate()
            except AttributeError as err:
                print(f'ExecutionProcessManager.terminate_pipeline_process: {err}')

    def set_pipeline_process(
        self,
        pipeline_run_id: int,
        proc: multiprocessing.Process,
    ):
        self.terminate_pipeline_process(pipeline_run_id)
        self.pipeline_processes[pipeline_run_id] = proc

    def has_block_process(self, pipeline_run_id: int, block_run_id: int):
        return (
            pipeline_run_id in self.block_processes
            and block_run_id in self.block_processes[pipeline_run_id]
            and self.block_processes[pipeline_run_id][block_run_id].is_alive()
        )

    def set_block_process(
        self,
        pipeline_run_id: int,
        block_run_id: int,
        proc: multiprocessing.Process,
    ):
        self.terminate_block_process(pipeline_run_id, block_run_id)
        if pipeline_run_id not in self.block_processes:
            self.block_processes[pipeline_run_id] = dict()
        self.block_processes[pipeline_run_id][block_run_id] = proc

    def terminate_block_process(self, pipeline_run_id: int, block_run_id: int) -> None:
        if self.has_block_process(pipeline_run_id, block_run_id):
            self.block_processes[pipeline_run_id][block_run_id].terminate()

    def clean_up_processes(self, include_child_processes=True):
        """
        Clean up inactive processes and terminate processes for cancelled pipeline runs
        """
        for pipeline_run_id in list(self.block_processes.keys()):
            block_run_procs = self.block_processes[pipeline_run_id]
            if not block_run_procs:
                del self.block_processes[pipeline_run_id]
                continue
            # TODO: Improve perf by batch fetching the pipeline runs
            pipeline_run = PipelineRun.query.get(pipeline_run_id)
            if pipeline_run.status == PipelineRun.PipelineRunStatus.CANCELLED:
                for block_run_id in list(block_run_procs.keys()):
                    proc = block_run_procs[block_run_id]
                    if proc.is_alive():
                        proc.terminate()
                    del block_run_procs[block_run_id]
            else:
                for block_run_id in list(block_run_procs.keys()):
                    proc = block_run_procs[block_run_id]
                    if proc and hasattr(proc, 'is_alive') and not proc.is_alive():
                        del block_run_procs[block_run_id]
            if not block_run_procs:
                del self.block_processes[pipeline_run_id]

        for pipeline_run_id in list(self.pipeline_processes.keys()):
            pipeline_run_proc = self.pipeline_processes[pipeline_run_id]
            if not pipeline_run_proc.is_alive():
                del self.pipeline_processes[pipeline_run_id]
            else:
                pipeline_run = PipelineRun.query.get(pipeline_run_id)
                if pipeline_run and pipeline_run.status == PipelineRun.PipelineRunStatus.CANCELLED:
                    pipeline_run_proc.terminate()
                    del self.pipeline_processes[pipeline_run_id]


execution_process_manager = ExecutionProcessManager()



from distutils.file_util import copy_file
from mage_ai.data_preparation.models.pipeline import Pipeline
from typing import Callable
import asyncio
import multiprocessing
import os
import shutil


class PipelineExecution:
    """
    Singleton class to manage the current pipeline execution running in
    the websocket. We need to track the state of this execution because
    it will be run in a seperate process, and the user can choose to cancel
    the execution.
    """
    def __init__(self):
        self.current_pipeline_process: multiprocessing.Process = None
        self.current_message_task = None
        self.previous_config_path: str = None


pipeline_execution = PipelineExecution()


def check_pipeline_process_status(
    pipeline: Pipeline,
    publish_message: Callable[..., None],
) -> None:
    current_process = pipeline_execution.current_pipeline_process
    is_alive = False
    if current_process:
        is_alive = current_process.is_alive()
    status = 'running' if is_alive else 'not running'

    publish_message(
        f'Pipeline {pipeline.uuid} is currently {status}.',
        metadata=dict(pipeline_uuid=pipeline.uuid),
    )


def set_current_pipeline_process(process: multiprocessing.Process) -> None:
    """
    Set the process that the current pipeline execution is running in.
    """
    pipeline_execution.current_pipeline_process = process


def set_current_message_task(task: asyncio.Task) -> None:
    """
    Set the task that current is processing messages from execution process.
    """
    pipeline_execution.current_message_task = task


def cancel_pipeline_execution(
    pipeline: Pipeline,
    publish_message: Callable[..., None],
    skip_publish_message: bool = False,
) -> None:
    """
    Cancel the current pipeline execution running in the saved
    process if the process is alive.
    """
    current_process = pipeline_execution.current_pipeline_process
    if current_process and current_process.is_alive():
        pipeline_execution.current_pipeline_process.terminate()

    if pipeline_execution.current_message_task:
        pipeline_execution.current_message_task.cancel()

    if not skip_publish_message:
        publish_message(
            'Pipeline execution cancelled... reverting state to previous iteration',
            execution_state='idle',
            metadata=dict(pipeline_uuid=pipeline.uuid),
        )
    config_path = pipeline_execution.previous_config_path
    if config_path is not None and os.path.isdir(config_path):
        copy_file(
            os.path.join(config_path, 'metadata.yaml'),
            os.path.join(pipeline.dir_path, 'metadata.yaml'),
        )
        delete_pipeline_copy_config(config_path)


def reset_execution_manager() -> None:
    """
    Reset state on the execution manager.
    """
    pipeline_execution.current_pipeline_process = None
    pipeline_execution.previous_config_path = None


def set_previous_config_path(path: str) -> None:
    """
    Save the path where we save the copy of the pipeline config before
    running the execution.
    """
    pipeline_execution.previous_config_path = path


def delete_pipeline_copy_config(path: str) -> None:
    """
    Delete the files for the copy of the pipeline config. This should
    only be used to delete the extra copy.
    """
    if os.path.isdir(path):
        shutil.rmtree(path)



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from typing import List


@dataclass
class EmailConfig(BaseConfig):
    smtp_host: str
    smtp_mail_from: str
    smtp_user: str
    smtp_password: str
    smtp_starttls: bool = True
    smtp_ssl: bool = False
    smtp_port: int = 587
    to_emails: List[str] = None

    @property
    def is_valid(self) -> bool:
        return len(self.to_emails or []) > 0



def number_string(version: int, number_of_numbers: int = 20) -> str:
    version_string = str(version)
    for i in range(number_of_numbers - len(version_string)):
        version_string = f'0{version_string}'

    return version_string



from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.io.export_utils import BadConversionError, PandasTypes
from mage_ai.io.sql import BaseSQL
from mage_ai.shared.utils import clean_name
from mysql.connector import connect
from mysql.connector.cursor import MySQLCursor
from pandas import DataFrame, Series
from typing import IO, List, Mapping, Union
import numpy as np


QUERY_ROW_LIMIT = 10_000_000


class MySQL(BaseSQL):
    def __init__(
        self,
        database: str,
        host: str,
        password: str,
        user: str,
        port: int = 3306,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            database=database,
            host=host,
            password=password,
            port=port or 3306,
            user=user,
            verbose=verbose,
            **kwargs,
        )

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'MySQL':
        return cls(
            database=config[ConfigKey.MYSQL_DATABASE],
            host=config[ConfigKey.MYSQL_HOST],
            password=config[ConfigKey.MYSQL_PASSWORD],
            port=config[ConfigKey.MYSQL_PORT],
            user=config[ConfigKey.MYSQL_USER],
        )

    def build_create_table_command(
        self,
        dtypes: Mapping[str, str],
        schema_name: str,
        table_name: str,
        unique_constraints: List[str] = [],
    ) -> str:
        query = []
        for cname in dtypes:
            query.append(f'`{clean_name(cname)}` {dtypes[cname]}')

        return f'CREATE TABLE {table_name} (' + ','.join(query) + ');'

    def open(self) -> None:
        with self.printer.print_msg('Opening connection to MySQL database'):
            self._ctx = connect(**self.settings)

    def table_exists(self, schema_name: str, table_name: str) -> bool:
        with self.conn.cursor() as cur:
            database_name = self.settings['database']
            cur.execute('\n'.join([
                'SELECT * FROM information_schema.tables ',
                f'WHERE table_schema = \'{database_name}\' AND table_name = \'{table_name}\'',
                'LIMIT 1',
            ]))
            return len(cur.fetchall()) >= 1

    def upload_dataframe(
        self,
        cursor: MySQLCursor,
        df: DataFrame,
        db_dtypes: List[str],
        dtypes: List[str],
        full_table_name: str,
        buffer: Union[IO, None] = None,
        **kwargs,
    ) -> None:
        values_placeholder = ', '.join(["%s" for i in range(len(df.columns))])
        values = []
        for i, row in df.iterrows():
            values.append(tuple(row))

        sql = f'INSERT INTO {full_table_name} VALUES ({values_placeholder})'
        cursor.executemany(sql, values)

    def get_type(self, column: Series, dtype: str) -> str:
        if dtype in (
            PandasTypes.MIXED,
            PandasTypes.UNKNOWN_ARRAY,
            PandasTypes.COMPLEX,
        ):
            raise BadConversionError(
                f'Cannot convert column \'{column.name}\' with data type \'{dtype}\' to '
                'a MySQL datatype.'
            )
        elif dtype in (PandasTypes.DATETIME, PandasTypes.DATETIME64):
            try:
                if column.dt.tz:
                    return 'TIMESTAMP'
            except AttributeError:
                pass
            return 'TIMESTAMP'
        elif dtype == PandasTypes.TIME:
            try:
                if column.dt.tz:
                    return 'TIME'
            except AttributeError:
                pass
            return 'TIME'
        elif dtype == PandasTypes.DATE:
            return 'DATE'
        elif dtype == PandasTypes.STRING:
            return 'CHAR(255)'
        elif dtype == PandasTypes.CATEGORICAL:
            return 'TEXT'
        elif dtype == PandasTypes.BYTES:
            return 'VARBINARY(255)'
        elif dtype in (PandasTypes.FLOATING, PandasTypes.DECIMAL, PandasTypes.MIXED_INTEGER_FLOAT):
            return 'DECIMAL'
        elif dtype == PandasTypes.INTEGER:
            max_int, min_int = column.max(), column.min()
            if np.int16(max_int) == max_int and np.int16(min_int) == min_int:
                return 'BIGINT'
            elif np.int32(max_int) == max_int and np.int32(min_int) == min_int:
                return 'BIGINT'
            else:
                return 'BIGINT'
        elif dtype == PandasTypes.BOOLEAN:
            return 'CHAR(52)'
        elif dtype in (PandasTypes.TIMEDELTA, PandasTypes.TIMEDELTA64, PandasTypes.PERIOD):
            return 'BIGINT'
        elif dtype == PandasTypes.EMPTY:
            return 'CHAR(255)'
        else:
            print(f'Invalid datatype provided: {dtype}')

        return 'CHAR(255)'

    def _enforce_limit(self, query: str, limit: int = QUERY_ROW_LIMIT) -> str:
        query = query.strip(';')

        return f"""
SELECT *
FROM (
    {query}
) a
LIMIT {limit}
"""



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.PipelineRunPresenter import PipelineRunPresenter
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db.models.oauth import Permission


class PipelineRunPolicy(BasePolicy):
    @property
    def entity(self):
        query = self.options.get('query', {})
        pipeline_uuid = query.get('pipeline_uuid', [None])
        if pipeline_uuid:
            pipeline_uuid = pipeline_uuid[0]
        if pipeline_uuid:
            return Permission.Entity.PIPELINE, pipeline_uuid

        if self.resource and self.resource.model:
            return Permission.Entity.PIPELINE, self.resource.model.pipeline_uuid

        return Permission.Entity.PROJECT, get_repo_path()


PipelineRunPolicy.allow_actions([
    constants.DETAIL,
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelineRunPolicy.allow_actions([
    constants.CREATE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelineRunPolicy.allow_read(PipelineRunPresenter.default_attributes + [
    'block_runs',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelineRunPolicy.allow_read(PipelineRunPresenter.default_attributes + [
    'block_runs',
    'block_runs_count',
    'pipeline_schedule_name',
    'pipeline_schedule_token',
    'pipeline_schedule_type',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelineRunPolicy.allow_read(PipelineRunPresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelineRunPolicy.allow_write([
    'backfill_id',
    'event_variables',
    'execution_date',
    'pipeline_schedule_id',
    'pipeline_uuid',
    'variables',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelineRunPolicy.allow_write([
    'from_block_uuid',
    'pipeline_run_action',
    'status',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelineRunPolicy.allow_query([
    'backfill_id',
    'order_by[]',
    'pipeline_uuid',
    'status',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())



from mage_ai.data_preparation.logging import LoggingConfig
from mage_ai.data_preparation.models.constants import LOGS_DIR
from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.repo_manager import RepoConfig, get_repo_config
from mage_ai.shared.array import find
import io
import logging
import os

MAX_LOG_FILE_SIZE = 20 * 1024 * 1024


class LoggerManager:
    def __init__(
        self,
        repo_path: str = None,
        logs_dir: str = None,
        pipeline_uuid: str = None,
        block_uuid: str = None,
        partition: str = None,
        repo_config: RepoConfig = None,
    ):
        self.repo_path = repo_path
        self.logs_dir = logs_dir
        self.pipeline_uuid = pipeline_uuid
        self.block_uuid = block_uuid
        self.partition = partition

        self.repo_config = repo_config or get_repo_config()
        logging_config = self.repo_config.logging_config if self.repo_config else dict()
        self.logging_config = LoggingConfig.load(config=logging_config)

        logger_name_parts = [self.pipeline_uuid]
        if self.partition is not None:
            logger_name_parts.append(self.partition)
        if self.block_uuid is not None:
            logger_name_parts.append(self.block_uuid)
        logger_name = '/'.join(logger_name_parts)

        self.logger = logging.getLogger(logger_name)

        self.log_level = logging.getLevelName(self.logging_config.level)
        self.logger.setLevel(self.log_level)

        self.formatter = logging.Formatter(
            '%(asctime)s %(message)s',
            '%Y-%m-%dT%H:%M:%S',
        )
        self.stream = None
        if not self.logger.handlers:
            if self.logging_config.destination_config:
                handler = self.create_stream_handler()
            else:
                log_filepath = self.get_log_filepath(create_dir=True)
                handler = logging.handlers.RotatingFileHandler(
                    log_filepath,
                    backupCount=10,
                    maxBytes=MAX_LOG_FILE_SIZE,
                )

            handler.setLevel(self.log_level)
            handler.setFormatter(self.formatter)
            self.logger.addHandler(handler)
        else:
            if self.logging_config.destination_config:
                stream_handler = \
                    find(lambda hr: hr.__class__ == logging.StreamHandler, self.logger.handlers)
                if stream_handler:
                    self.stream = stream_handler.stream

    def create_stream_handler(self):
        self.stream = io.StringIO()
        return logging.StreamHandler(self.stream)

    def output_logs_to_destination(self):
        pass

    def create_log_filepath_dir(self, path):
        if not os.path.exists(path):
            os.makedirs(path)

    def get_log_filepath_prefix(self):
        logs_dir = self.logs_dir or self.repo_config.variables_dir

        return os.path.join(
            logs_dir,
            'pipelines',
            self.pipeline_uuid,
            LOGS_DIR,
            self.partition or '',
        )

    def get_log_filepath(self, create_dir: bool = False):
        if self.pipeline_uuid is None:
            raise Exception('Please specify a pipeline uuid in your logger.')

        prefix = self.get_log_filepath_prefix()

        if create_dir:
            self.create_log_filepath_dir(prefix)

        if self.block_uuid is None:
            log_filepath = os.path.join(prefix, 'pipeline.log')
        else:
            log_filepath = os.path.join(prefix, f'{self.block_uuid}.log')
        return log_filepath

    def get_logs(self):
        file = File.from_path(self.get_log_filepath())
        return file.to_dict(include_content=True)

    async def get_logs_async(self):
        file = File.from_path(self.get_log_filepath())
        return await file.to_dict_async(include_content=True)



import json
import os
import sys
from typing import List, Union

import typer
from click import Context
from rich import print
from typer.core import TyperGroup

from mage_ai.cli.utils import parse_runtime_variables


class OrderCommands(TyperGroup):
    def list_commands(self, ctx: Context):
        """Return list of commands in the order they appear."""
        return list(self.commands)


app = typer.Typer(
    cls=OrderCommands,
)


@app.command()
def init(
    project_path: str = typer.Argument(..., help='path of the Mage project to be created.'),
    # project_type: Union[str, None] = typer.Option(
    #     'standalone', help='type of project to create, options are main, sub, or standalone'
    # ),
):
    """
    Initialize Mage project.
    """
    from mage_ai.data_preparation.repo_manager import init_repo

    repo_path = os.path.join(os.getcwd(), project_path)
    init_repo(repo_path)
    print(f'Initialized Mage project at {repo_path}')


@app.command()
def start(
    project_path: str = typer.Argument(os.getcwd(), help='path of the Mage project to be loaded.'),
    host: str = typer.Option('localhost', help='specify the host.'),
    port: str = typer.Option('6789', help='specify the port.'),
    manage_instance: str = typer.Option('0', help=''),
    dbt_docs_instance: str = typer.Option('0', help=''),
):
    """
    Start Mage server and UI.
    """
    from mage_ai.data_preparation.repo_manager import set_repo_path

    # Set repo_path before intializing the DB so that we can get correct db_connection_url
    project_path = os.path.abspath(project_path)
    set_repo_path(project_path)

    from mage_ai.server.server import start_server

    start_server(
        host=host,
        port=port,
        project=project_path,
        manage=manage_instance == "1",
        dbt_docs=dbt_docs_instance == "1",
    )


@app.command()
def run(
    project_path: str = typer.Argument(
        ..., help='path of the Mage project that contains the pipeline.'
    ),
    pipeline_uuid: str = typer.Argument(
        ..., help='uuid of the pipeline to be run.'
    ),
    test: bool = typer.Option(
        False, help='specify if tests should be run.'
    ),
    block_uuid: Union[str, None] = typer.Option(
        None, help='uuid of the block to be run.'
    ),
    execution_partition: Union[str, None] = typer.Option(
        None, help=''
    ),
    executor_type: Union[str, None] = typer.Option(
        None, help=''
    ),
    callback_url: Union[str, None] = typer.Option(
        None, help=''
    ),
    block_run_id: Union[int, None] = typer.Option(
        None, help=''
    ),
    pipeline_run_id: Union[int, None] = typer.Option(
        None, help=''
    ),
    runtime_vars: Union[List[str], None] = typer.Option(
        None, help='specify runtime variables. These will overwrite the pipeline global variables.'
    ),
    skip_sensors: bool = typer.Option(
        False, help='specify if the sensors should be skipped.'
    ),
    template_runtime_configuration: Union[str, None] = typer.Option(
        None, help='runtime configuration of data integration block runs.'
    ),
):
    """
    Run pipeline.
    """
    from mage_ai.data_preparation.repo_manager import set_repo_path

    # Set repo_path before intializing the DB so that we can get correct db_connection_url
    project_path = os.path.abspath(project_path)
    set_repo_path(project_path)

    import sentry_sdk

    from mage_ai.data_preparation.executors.executor_factory import ExecutorFactory
    from mage_ai.data_preparation.models.pipeline import Pipeline
    from mage_ai.data_preparation.variable_manager import get_global_variables
    from mage_ai.orchestration.db import db_connection
    from mage_ai.settings import SENTRY_DSN, SENTRY_TRACES_SAMPLE_RATE
    from mage_ai.shared.hash import merge_dict

    sentry_dsn = SENTRY_DSN
    if sentry_dsn:
        sentry_sdk.init(
            sentry_dsn,
            traces_sample_rate=SENTRY_TRACES_SAMPLE_RATE,
        )

    runtime_variables = dict()
    if runtime_vars is not None:
        runtime_variables = parse_runtime_variables(runtime_vars)

    sys.path.append(os.path.dirname(project_path))
    pipeline = Pipeline.get(pipeline_uuid, repo_path=project_path)

    default_variables = get_global_variables(pipeline_uuid)
    global_vars = merge_dict(default_variables, runtime_variables)

    db_connection.start_session()

    if template_runtime_configuration is not None:
        template_runtime_configuration = json.loads(template_runtime_configuration)

    if block_uuid is None:
        ExecutorFactory.get_pipeline_executor(
            pipeline,
            execution_partition=execution_partition,
            executor_type=executor_type,
        ).execute(
            analyze_outputs=False,
            global_vars=global_vars,
            pipeline_run_id=pipeline_run_id,
            run_sensors=not skip_sensors,
            run_tests=test,
            update_status=False,
        )
    else:
        ExecutorFactory.get_block_executor(
            pipeline,
            block_uuid,
            execution_partition=execution_partition,
            executor_type=executor_type,
        ).execute(
            analyze_outputs=False,
            block_run_id=block_run_id,
            callback_url=callback_url,
            global_vars=global_vars,
            pipeline_run_id=pipeline_run_id,
            template_runtime_configuration=template_runtime_configuration,
            update_status=False,
        )
    print('Pipeline run completed.')


@app.command()
def clean_cached_variables(
    project_path: str = typer.Argument(
        ..., help='path of the Mage project to clean variables.'
    ),
    pipeline_uuid: str = typer.Option(
        None, help='uuid of the pipeline to clean.'
    ),
):
    from mage_ai.data_preparation.repo_manager import set_repo_path

    project_path = os.path.abspath(project_path)
    set_repo_path(project_path)

    from mage_ai.data_preparation.variable_manager import clean_variables

    clean_variables(pipeline_uuid=pipeline_uuid)


@app.command()
def create_spark_cluster(
    project_path: str = typer.Argument(
        ..., help='path of the Mage project that contains the EMR config.'
    ),
):
    """
    Create EMR cluster for Mage project.
    """
    from mage_ai.services.aws.emr.launcher import create_cluster

    project_path = os.path.abspath(project_path)
    create_cluster(project_path)


if __name__ == '__main__':
    app()



from mage_ai.services.slack.config import SlackConfig
import json
import requests


def send_slack_message(config: SlackConfig, message: str) -> None:
    requests.post(
        config.webhook_url,
        json.dumps(dict(text=message)),
    )



from dataclasses import dataclass
from enum import Enum


class SerializationMethod(str, Enum):
    AVRO = 'AVRO'
    JSON = 'JSON'
    PROTOBUF = 'PROTOBUF'
    RAW_VALUE = 'RAW_VALUE'


@dataclass
class SerDeConfig:
    serialization_method: SerializationMethod
    schema_classpath: str = None            # Used by PROTOBUF
    schema_registry_url: str = None         # Used by AVRO
    schema_registry_username: str = None    # Used by AVRO
    schema_registry_password: str = None    # Used by AVRO



from mage_ai.data_cleaner.transformer_actions.udf.base import BaseUDF
import pandas as pd


class FormattedDate(BaseUDF):
    def execute(self):
        return pd.to_datetime(
            self.df[self.arguments[0]],
        ).dt.strftime(self.options['format'])



from mage_ai.api.presenters.BasePresenter import BasePresenter


class RolePresenter(BasePresenter):
    default_attributes = [
        'id',
        'created_at',
        'name',
        'permissions',
        'updated_at',
    ]

    async def present(self, **kwargs):
        return self.model.to_dict(include_attributes=self.default_attributes)



from mage_ai.streaming.sources.google_cloud_pubsub import GoogleCloudPubSubSource
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class GoogleCloudPubSubTests(TestCase):
    def test_init(self):
        with patch.object(GoogleCloudPubSubSource, 'init_client') as mock_init_client:
            GoogleCloudPubSubSource(dict(
                connector_type='test_connector_type',
                project_id='test_project_id',
                topic_id='test_topic_id',
                subscription_id='test_subscription_id',
                timeout=5,
                batch_size=100,
            ))
            mock_init_client.assert_called_once()

    def test_init_invalid_config(self):
        with patch.object(GoogleCloudPubSubSource, 'init_client') as mock_init_client:
            with self.assertRaises(Exception) as context:
                GoogleCloudPubSubSource(dict(
                    connector_type='test_connector_type',
                    project_id='test_project_id',
                    topic_id='test_topic_id',
                ))
            self.assertTrue(
                '__init__() missing 1 required positional argument: \'subscription_id\''
                in str(context.exception),
            )
            self.assertEqual(mock_init_client.call_count, 0)



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.cluster_manager.constants import (
    ECS_CLUSTER_NAME,
    ECS_CONTAINER_NAME,
    ECS_TASK_DEFINITION,
    GCP_PATH_TO_KEYFILE,
    GCP_PROJECT_ID,
    GCP_REGION,
    KUBE_NAMESPACE,
    KUBE_SERVICE_ACCOUNT_NAME,
    KUBE_STORAGE_CLASS_NAME,
)
from mage_ai.data_preparation.repo_manager import (
    get_project_type,
    get_repo_config,
    get_repo_path,
    ProjectType,
)
from mage_ai.data_preparation.shared.constants import MANAGE_ENV_VAR
from mage_ai.orchestration.db import safe_db_query
from mage_ai.server.api.clusters import ClusterType
from mage_ai.server.logger import Logger
from typing import Dict, List
import os
import shutil
import yaml

logger = Logger().new_server_logger(__name__)


class WorkspaceResource(GenericResource):
    @classmethod
    @safe_db_query
    def collection(self, query_arg, meta, user, **kwargs):
        cluster_type = self.verify_project()
        if not cluster_type:
            cluster_type = query_arg.get('cluster_type', [None])
            if cluster_type:
                cluster_type = cluster_type[0]

        instances = self.get_instances(cluster_type)
        instance_map = {
            instance.get('name'): instance
            for instance in instances
        }

        repo_path = get_repo_path()
        projects_folder = os.path.join(repo_path, 'projects')
        if get_project_type() == ProjectType.MAIN:
            projects = [name for name in os.listdir(projects_folder) if os.path.isdir(name)]
        else:
            projects = [instance.get('name') for instance in instances]

        workspaces = [
            dict(
                name=project,
                cluster_type=cluster_type,
                instance=instance_map[project],
            )
            for project in projects
            if project in instance_map
        ]

        return self.build_result_set(workspaces, user, **kwargs)

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        cluster_type = self.verify_project(pk)
        if not cluster_type:
            query = kwargs.get('query', {})
            cluster_type = query.get('cluster_type')[0]

        instances = self.get_instances(cluster_type)
        instance_map = {
            instance.get('name'): instance
            for instance in instances
        }

        return self(dict(
            name=pk,
            cluster_type=cluster_type,
            instance=instance_map[pk],
        ), user, **kwargs)

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        cluster_type = self.verify_project()
        if not cluster_type:
            cluster_type = payload.get('cluster_type')

        if cluster_type == ClusterType.K8S:
            from mage_ai.cluster_manager.kubernetes.workload_manager import WorkloadManager
            name = payload.get('name')
            namespace = payload.get('namespace', os.getenv(KUBE_NAMESPACE))
            storage_class_name = payload.get(
                'storage_class_name',
                os.getenv(KUBE_STORAGE_CLASS_NAME),
            )
            service_account_name = payload.get(
                'service_account_name',
                os.getenv(KUBE_SERVICE_ACCOUNT_NAME),
            )
            container_config_yaml = payload.get('container_config')
            container_config = None
            if container_config_yaml:
                container_config = yaml.full_load(container_config_yaml)

            k8s_workload_manager = WorkloadManager(namespace)
            k8s_workload_manager.create_stateful_set(
                name,
                container_config=container_config,
                service_account_name=service_account_name,
                storage_class_name=storage_class_name,
            )
        elif cluster_type == ClusterType.ECS:
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager
            name = payload.get('name')
            cluster_name = payload.get('cluster_name', os.getenv(ECS_CLUSTER_NAME))
            task_definition = payload.get(
                'task_definition',
                os.getenv(ECS_TASK_DEFINITION),
            )
            container_name = payload.get(
                'container_name',
                os.getenv(ECS_CONTAINER_NAME),
            )

            ecs_instance_manager = EcsTaskManager(cluster_name)

            ecs_instance_manager.create_task(
                name,
                task_definition,
                container_name,
            )
        elif cluster_type == ClusterType.CLOUD_RUN:
            from mage_ai.cluster_manager.gcp.cloud_run_service_manager \
                import CloudRunServiceManager
            name = payload.get('name')
            project_id = payload.get('project_id', os.getenv(GCP_PROJECT_ID))
            path_to_credentials = payload.get(
                'path_to_credentials',
                os.getenv('path_to_keyfile')
            )
            region = payload.get('region', os.getenv('GCP_REGION'))

            cloud_run_service_manager = CloudRunServiceManager(
                project_id,
                path_to_credentials,
                region=region
            )

            cloud_run_service_manager.create_service(name)

        return self(dict(success=True), user, **kwargs)

    def update(self, payload, **kwargs):
        cluster_type = self.model.get('cluster_type')
        instance_name = self.model.get('name')
        if cluster_type == 'ecs':
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager
            task_arn = payload.get('task_arn')
            cluster_name = payload.get('cluster_name', os.getenv(ECS_CLUSTER_NAME))
            task_definition = payload.get(
                'task_definition',
                os.getenv(ECS_TASK_DEFINITION),
            )
            container_name = payload.get('container_name', os.getenv(ECS_CONTAINER_NAME))

            action = payload.get('action')

            ecs_instance_manager = EcsTaskManager(cluster_name)
            if action == 'stop':
                ecs_instance_manager.stop_task(task_arn)
            elif action == 'resume':
                ecs_instance_manager.create_task(
                    instance_name,
                    task_definition,
                    container_name,
                )

        return self

    def delete(self, **kwargs):
        cluster_type = self.model.get('cluster_type')
        workspace_name = self.model.get('name')
        instance = self.model.get('instance')

        repo_path = get_repo_path()
        project_folder = os.path.join(repo_path, 'projects', workspace_name)

        if get_project_type() == ProjectType.MAIN:
            shutil.rmtree(project_folder)
        if cluster_type == 'ecs':
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager
            task_arn = instance.get('task_arn')
            cluster_name = os.getenv(ECS_CLUSTER_NAME)

            ecs_instance_manager = EcsTaskManager(cluster_name)
            ecs_instance_manager.delete_task(workspace_name, task_arn)

        return self

    @classmethod
    def verify_project(self, subproject: str = None) -> str:
        project_type = get_project_type()
        if project_type != ProjectType.MAIN and os.getenv(MANAGE_ENV_VAR) != '1':
            error = ApiError.RESOURCE_ERROR.copy()
            error.update(message='This project is ineligible for workspace management.')
            raise ApiError(error)

        if project_type == ProjectType.MAIN and subproject:
            repo_path = get_repo_path()
            projects_folder = os.path.join(repo_path, 'projects')
            projects = [name for name in os.listdir(projects_folder) if os.path.isdir(name)]
            if subproject not in projects:
                error = ApiError.RESOURCE_NOT_FOUND.copy()
                error.update(message=f'Project {subproject} was not found.')
                raise ApiError(error)

        if project_type == ProjectType.MAIN:
            return get_repo_config().cluster_type

    @classmethod
    def get_instances(self, cluster_type: str) -> List[Dict]:
        instances = []
        if cluster_type == ClusterType.K8S:
            from mage_ai.cluster_manager.kubernetes.workload_manager import WorkloadManager
            namespace = os.getenv(KUBE_NAMESPACE)
            workload_manager = WorkloadManager(namespace)

            instances = workload_manager.list_services()
        elif cluster_type == ClusterType.ECS:
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager

            try:
                cluster_name = os.getenv(ECS_CLUSTER_NAME)
                ecs_instance_manager = EcsTaskManager(cluster_name)
                instances = ecs_instance_manager.list_tasks()
            except Exception as e:
                error = ApiError.RESOURCE_ERROR.copy()
                error.update(message=str(e))
                raise ApiError(error)
        elif cluster_type == ClusterType.CLOUD_RUN:
            from mage_ai.cluster_manager.gcp.cloud_run_service_manager import CloudRunServiceManager
            project_id = os.getenv(GCP_PROJECT_ID)
            path_to_credentials = os.getenv(GCP_PATH_TO_KEYFILE)
            region = os.getenv(GCP_REGION)
            cloud_run_service_manager = CloudRunServiceManager(
                project_id,
                path_to_credentials,
                region=region
            )

            instances = cloud_run_service_manager.list_services()

        return instances



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.DatabaseResource import DatabaseResource
from mage_ai.api.utils import get_access_for_roles
from mage_ai.authentication.oauth2 import encode_token, generate_access_token
from mage_ai.authentication.passwords import (
    create_bcrypt_hash,
    generate_salt,
    verify_password,
)
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.oauth import Permission, Role, User
from mage_ai.shared.hash import extract, ignore_keys
from mage_ai.usage_statistics.logger import UsageStatisticLogger


class UserResource(DatabaseResource):
    model_class = User

    def __init__(self, model, current_user, **kwargs):
        super().__init__(model, current_user, **kwargs)
        self.group = None

    @classmethod
    @safe_db_query
    def collection(self, query_arg, meta, user, **kwargs):
        results = (
            User.
            query.
            order_by(User.username.asc())
        )

        if user.is_admin:
            results = list(filter(lambda user: user.project_access & 3 == 0, results))

        return results

    @classmethod
    @safe_db_query
    async def create(self, payload, user, **kwargs):
        email = payload.get('email')
        password = payload.get('password')
        password_confirmation = payload.get('password_confirmation')
        username = payload.get('username')

        error = ApiError.RESOURCE_INVALID.copy()

        role_ids = payload.get('roles_new', [])
        roles_new = self.check_roles(role_ids)

        missing_values = []
        if len(roles_new) == 0:
            missing_values.append('roles')

        payload['roles_new'] = roles_new

        for key in ['email', 'password']:
            if not payload.get(key):
                missing_values.append(key)

        if len(roles_new) == 0:
            missing_values.append('roles')

        if len(missing_values) >= 1:
            error.update(
                {'message': 'Missing required values: {}.'.format(', '.join(missing_values))})
            raise ApiError(error)

        if email:
            user = User.query.filter(User.email == email).first()
            if user:
                error.update(
                    {'message': f'Account with same email is already taken: {email}.'})
                raise ApiError(error)
        if username:
            user = User.query.filter(User.username == username).first()
            if user:
                error.update(
                    {'message': f'Account with same username is already taken: {username}.'})
                raise ApiError(error)

        if len(password) < 8:
            error.update(
                {'message': 'Password must be 8 characters or longer.'})
            raise ApiError(error)

        if password != password_confirmation:
            error.update(
                {'message': 'Password and password confirmation do not match.'})
            raise ApiError(error)

        password_salt = generate_salt()
        payload['email'] = email
        payload['password_hash'] = create_bcrypt_hash(password, password_salt)
        payload['password_salt'] = password_salt

        resource = super().create(extract(payload, [
            'avatar',
            'email',
            'first_name',
            'last_name',
            'password_hash',
            'password_salt',
            'roles',
            'roles_new',
            'username',
        ]), user, **kwargs)

        if 'oauth_client' in kwargs:
            oauth_token = generate_access_token(
                resource.model, kwargs['oauth_client'])
            resource.model_options['oauth_token'] = oauth_token

        async def _create_callback(resource):
            await UsageStatisticLogger().users_impression()

        self.on_create_callback = _create_callback

        return resource

    @safe_db_query
    def update(self, payload, **kwargs):
        error = ApiError.RESOURCE_INVALID.copy()

        if 'roles_new' in payload:
            role_ids = payload.get('roles_new', [])
            roles_new = self.check_roles(role_ids)

            missing_values = []
            if len(roles_new) == 0:
                missing_values.append('roles')

            payload['roles_new'] = roles_new

            if len(missing_values) >= 1:
                error.update(
                    {'message': 'Missing required values: {}.'.format(', '.join(missing_values))})
                raise ApiError(error)

            access = get_access_for_roles(roles_new, Permission.Entity.PROJECT, get_repo_path())

            if self.current_user.is_admin:
                if self.owner:
                    error.update(
                        {'message': 'Admins cannot update users who are Owners.'})
                    raise ApiError(error)
                elif self.is_admin and self.current_user.id != self.id:
                    error.update(
                        {'message': 'Admins cannot update users who are Admins.'})
                    raise ApiError(error)
                elif payload.get('roles') and int(payload.get('roles')) & 1 != 0 or \
                        access & Permission.Access.ADMIN != 0:
                    error.update(
                        {'message': 'Admins cannot make other users Admins.'})
                    raise ApiError(error)
                elif access & Permission.Access.OWNER != 0:
                    error.update(
                        {'message': 'Admins cannot make other users Owners.'})
                    raise ApiError(error)

        password = payload.get('password')
        if password:
            password_current = payload.get('password_current')
            password_confirmation = payload.get('password_confirmation')

            if self.current_user.id == self.id or \
                    (not self.current_user.owner and self.current_user.roles & 1 == 0):
                if not password_current or not verify_password(
                    password_current,
                    self.password_hash,
                ):
                    error.update(
                        {'message': 'Current password is incorrect.'})
                    raise ApiError(error)

            if len(password) < 8:
                error.update(
                    {'message': 'Password must be 8 characters or longer.'})
                raise ApiError(error)

            if password != password_confirmation:
                error.update(
                    {'message': 'Password and password confirmation do not match.'})
                raise ApiError(error)

            password_salt = generate_salt()
            payload['password_hash'] = create_bcrypt_hash(password, password_salt)
            payload['password_salt'] = password_salt

        return super().update(ignore_keys(payload, [
            'password',
            'password_confirmation',
            'password_current',
        ]), **kwargs)

    @safe_db_query
    def token(self):
        oauth_token = self.model_options.get('oauth_token')
        if oauth_token:
            return encode_token(oauth_token.token, oauth_token.expires)

    @classmethod
    @safe_db_query
    def check_roles(self, role_ids):
        missing_ids = []
        roles_new = []
        for role_id in role_ids:
            role = Role.query.get(int(role_id))
            if role is None:
                missing_ids.append(role_id)
            else:
                roles_new.append(role)

        if len(missing_ids) > 0:
            error = ApiError.RESOURCE_INVALID.copy()
            error.update(
                {'message': f'Roles with ids: {missing_ids} do not exist'})
            raise ApiError(error)

        return roles_new



from mage_ai.streaming.constants import SinkType
from typing import Dict


class SinkFactory:
    @classmethod
    def get_sink(self, config: Dict, **kwargs):
        connector_type = config['connector_type']
        if connector_type == SinkType.AMAZON_S3:
            from mage_ai.streaming.sinks.amazon_s3 import AmazonS3Sink

            return AmazonS3Sink(config, **kwargs)
        elif connector_type == SinkType.DUMMY:
            from mage_ai.streaming.sinks.dummy import DummySink

            return DummySink(config, **kwargs)
        elif connector_type == SinkType.KAFKA:
            from mage_ai.streaming.sinks.kafka import KafkaSink

            return KafkaSink(config, **kwargs)
        elif connector_type == SinkType.KINESIS:
            from mage_ai.streaming.sinks.kinesis import KinesisSink

            return KinesisSink(config, **kwargs)
        elif connector_type == SinkType.MONGODB:
            from mage_ai.streaming.sinks.mongodb import MongoDbSink

            return MongoDbSink(config, **kwargs)
        elif connector_type == SinkType.OPENSEARCH:
            from mage_ai.streaming.sinks.opensearch import OpenSearchSink

            return OpenSearchSink(config, **kwargs)
        raise Exception(
            f'Ingesting data to {connector_type} is not supported in streaming pipelines yet.',
        )



import os

from mage_ai.orchestration.notification.config import AlertOn, NotificationConfig
from mage_ai.services.email.email import send_email
from mage_ai.services.google_chat.google_chat import send_google_chat_message
from mage_ai.services.opsgenie.opsgenie import send_opsgenie_alert
from mage_ai.services.slack.slack import send_slack_message
from mage_ai.services.teams.teams import send_teams_message
from mage_ai.settings import MAGE_PUBLIC_HOST


class NotificationSender:
    def __init__(self, config: NotificationConfig):
        self.config = config

    def send(
        self,
        title: str = None,
        summary: str = None,
        details: str = None,
    ) -> None:
        """Send messages to the notification channels.

        Args:
            title (str, optional): Short sentence, used as title (e.g. Email subject)
            summary (str, optional): Mid-length sentences, used as the summary of the message.
            details (str, optional): Long message, used as the body of the message (e.g. Email body)
        """
        if summary is None:
            return
        if self.config.slack_config is not None and self.config.slack_config.is_valid:
            send_slack_message(self.config.slack_config, details or summary)

        if self.config.teams_config is not None and self.config.teams_config.is_valid:
            send_teams_message(self.config.teams_config, summary)

        if self.config.google_chat_config is not None and self.config.google_chat_config.is_valid:
            send_google_chat_message(self.config.google_chat_config, summary)

        if self.config.email_config is not None and title is not None:
            send_email(
                self.config.email_config,
                subject=title,
                message=details or summary,
            )

        if self.config.opsgenie_config is not None and self.config.opsgenie_config.is_valid:
            send_opsgenie_alert(
                self.config.opsgenie_config,
                message=title,
                description=details or summary,
            )

    def send_pipeline_run_success_message(self, pipeline, pipeline_run) -> None:
        if AlertOn.PIPELINE_RUN_SUCCESS in self.config.alert_on:
            message = (
                f'Successfully ran Pipeline `{pipeline.uuid}` '
                f'with Trigger {pipeline_run.pipeline_schedule.id} '
                f'`{pipeline_run.pipeline_schedule.name}` '
                f'at execution time `{pipeline_run.execution_date}`.'
            )
            email_content = f'{message}\n'
            if os.getenv('ENV') != 'production':
                email_content += f'Open {self.__pipeline_run_url(pipeline, pipeline_run)} '\
                                  'to check pipeline run results and logs.'
            self.send(
                title=f'Successfully ran Pipeline {pipeline.uuid}',
                summary=message,
                details=email_content,
            )

    def send_pipeline_run_failure_message(
        self,
        pipeline,
        pipeline_run,
        message: str = None,
    ) -> None:
        if AlertOn.PIPELINE_RUN_FAILURE in self.config.alert_on:
            message = message or (
                f'Failed to run Pipeline `{pipeline.uuid}` '
                f'with Trigger {pipeline_run.pipeline_schedule.id} '
                f'`{pipeline_run.pipeline_schedule.name}` '
                f'at execution time `{pipeline_run.execution_date}`.'
            )
            email_content = f'{message}\n'
            if os.getenv('ENV') != 'production' or MAGE_PUBLIC_HOST != 'http://localhost:6789':
                """
                Include the URL for the following cases
                1. Dev environment: Use the default localhost as host in URL
                2. Production environment: If MAGE_PUBLIC_HOST is set, use it as host.
                """
                email_content += f'Open {self.__pipeline_run_url(pipeline, pipeline_run)} '\
                                 'to check pipeline run results and logs.'
            self.send(
                title=f'Failed to run Mage pipeline {pipeline.uuid}',
                summary=message,
                details=email_content,
            )

    def send_pipeline_run_sla_passed_message(self, pipeline, pipeline_run) -> None:
        if AlertOn.PIPELINE_RUN_PASSED_SLA in self.config.alert_on:
            message = (
                f'SLA passed for pipeline `{pipeline.uuid}` '
                f'with Trigger {pipeline_run.pipeline_schedule.id} '
                f'`{pipeline_run.pipeline_schedule.name}` '
                f'at execution time `{pipeline_run.execution_date}`.'
            )
            email_content = f'{message}\n'
            if os.getenv('ENV') != 'production':
                email_content += f'Open {self.__pipeline_run_url(pipeline, pipeline_run)} '\
                                  'to check pipeline run results and logs.'
            self.send(
                title=f'SLA passed for Mage pipeline {pipeline.uuid}',
                summary=message,
                details=email_content,
            )

    @staticmethod
    def __pipeline_run_url(pipeline, pipeline_run):
        return f'{MAGE_PUBLIC_HOST}/pipelines/{pipeline.uuid}/triggers/'\
               f'{pipeline_run.pipeline_schedule_id}'



from mage_ai.api.presenters.BasePresenter import BasePresenter


class VariablePresenter(BasePresenter):
    default_attributes = [
        'block',
        'name',
        'pipeline',
        'value',
        'variables',
    ]



from io import StringIO
from mage_ai.io.base import BaseSQLConnection, ExportWritePolicy, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader
from mage_ai.io.export_utils import (
    clean_df_for_export,
    gen_table_creation_query,
    infer_dtypes,
    PandasTypes,
)
from pandas import DataFrame, read_sql, Series
from typing import Any, Dict, IO, List, Mapping, Union
import warnings


class BaseSQL(BaseSQLConnection):
    @classmethod
    def with_config(cls, config: BaseConfigLoader):
        """
        Initializes SQL loader from configuration loader

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        raise Exception('Subclasses must override this method.')

    def get_type(self, column: Series, dtype: str) -> str:
        """
        Maps pandas Data Frame column to SQL type

        Args:
            series (Series): Column to map
            dtype (str): Pandas data type of this column

        Raises:
            ConversionError: Returned if this type cannot be converted to a SQL data type

        Returns:
            str: SQL data type for this column
        """
        raise Exception('Subclasses must override this method.')

    def build_create_table_command(
        self,
        dtypes: Mapping[str, str],
        schema_name: str,
        table_name: str,
        unique_constraints: List[str] = [],
    ) -> str:
        return gen_table_creation_query(
            dtypes,
            schema_name,
            table_name,
            unique_constraints=unique_constraints,
        )

    def build_create_table_as_command(
        self,
        table_name: str,
        query_string: str,
    ) -> str:
        return 'CREATE TABLE {} AS\n{}'.format(
            table_name,
            query_string,
        )

    def default_database(self) -> str:
        return None

    def default_schema(self) -> str:
        return None

    def open(self) -> None:
        """
        Opens a connection to the SQL database specified by the parameters.
        """
        raise Exception('Subclasses must override this method.')

    def table_exists(self, schema_name: str, table_name: str) -> bool:
        """
        Returns whether the specified table exists.

        Args:
            schema_name (str): Name of the schema the table belongs to.
            table_name (str): Name of the table to check existence of.

        Returns:
            bool: True if the table exists, else False.
        """
        raise Exception('Subclasses must override this method.')

    def upload_dataframe(
        self,
        cursor,
        df: DataFrame,
        db_dtypes: List[str],
        dtypes: List[str],
        full_table_name: str,
        buffer: Union[IO, None] = None
    ) -> None:
        raise Exception('Subclasses must override this method.')

    def execute(self, query_string: str, **query_vars) -> None:
        """
        Sends query to the connected database.

        Args:
            query_string (str): SQL query string to apply on the connected database.
            query_vars: Variable values to fill in when using format strings in query.
        """
        with self.printer.print_msg(f'Executing query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            with self.conn.cursor() as cur:
                cur.execute(query_string, **query_vars)

    def execute_queries(
        self,
        queries: List[str],
        query_variables: List[Dict] = None,
        commit: bool = False,
        fetch_query_at_indexes: List[bool] = None,
    ) -> List:
        results = []

        with self.conn.cursor() as cursor:
            for idx, query in enumerate(queries):
                variables = query_variables[idx] \
                                if query_variables and idx < len(query_variables) \
                                else {}
                query = self._clean_query(query)

                if fetch_query_at_indexes and idx < len(fetch_query_at_indexes) and \
                        fetch_query_at_indexes[idx]:
                    result = self.fetch_query(
                        cursor,
                        query,
                    )
                else:
                    result = cursor.execute(query, **variables)

                results.append(result)

        if commit:
            self.conn.commit()

        return results

    def fetch_query(self, cursor, query: str) -> Any:
        return read_sql(query, self.conn)

    def load(
        self,
        query_string: str,
        limit: int = QUERY_ROW_LIMIT,
        display_query: Union[str, None] = None,
        verbose: bool = True,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from the connected database into a Pandas data frame based on the query given.
        This will fail if the query returns no data from the database. This function will load at
        maximum 10,000,000 rows of data. To operate on more data, consider performing data
        transformations in warehouse.

        Args:
            query_string (str): Query to execute on the database.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults
                to 10,000,000.
            **kwargs: Additional query parameters.

        Returns:
            DataFrame: The data frame corresponding to the data returned by the given query.
        """
        print_message = 'Loading data'
        if verbose:
            print_message += ' with query'

            if display_query:
                for line in display_query.split('\n'):
                    print_message += f'\n{line}'
            else:
                print_message += f'\n\n{query_string}\n\n'

        query_string = self._clean_query(query_string)

        with self.printer.print_msg(print_message):
            warnings.filterwarnings('ignore', category=UserWarning)

            return read_sql(
                self._enforce_limit(query_string, limit),
                self.conn,
                **kwargs,
            )

    def export(
        self,
        df: DataFrame,
        schema_name: str,
        table_name: str,
        if_exists: ExportWritePolicy = ExportWritePolicy.REPLACE,
        index: bool = False,
        verbose: bool = True,
        query_string: Union[str, None] = None,
        drop_table_on_replace: bool = False,
        cascade_on_drop: bool = False,
        allow_reserved_words: bool = False,
        unique_conflict_method: str = None,
        unique_constraints: List[str] = None,
    ) -> None:
        """
        Exports dataframe to the connected database from a Pandas data frame. If table doesn't
        exist, the table is automatically created. If the schema doesn't exist, the schema is
        also created.

        Args:
            schema_name (str): Name of the schema of the table to export data to.
            table_name (str): Name of the table to insert rows from this data frame into.
            if_exists (ExportWritePolicy): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table. In this case the schema must
                                match the original table.
            Defaults to `'replace'`.
            index (bool): If true, the data frame index is also exported alongside the table.
                            Defaults to False.
            **kwargs: Additional query parameters.
        """

        if type(df) is dict:
            df = DataFrame([df])
        elif type(df) is list:
            df = DataFrame(df)

        if schema_name:
            full_table_name = f'{schema_name}.{table_name}'
        else:
            full_table_name = table_name

        if not query_string:
            if index:
                df = df.reset_index()

            dtypes = infer_dtypes(df)
            df = clean_df_for_export(df, self.clean, dtypes)

            # Clean column names
            col_mapping = {col: self._clean_column_name(
                                        col,
                                        allow_reserved_words=allow_reserved_words)
                           for col in df.columns}
            df = df.rename(columns=col_mapping)
            dtypes = infer_dtypes(df)

        def __process():
            buffer = StringIO()
            table_exists = self.table_exists(schema_name, table_name)

            with self.conn.cursor() as cur:
                if schema_name:
                    cur.execute(f'CREATE SCHEMA IF NOT EXISTS {schema_name};')

                should_create_table = not table_exists

                if table_exists:
                    if ExportWritePolicy.FAIL == if_exists:
                        raise ValueError(
                            f'Table \'{full_table_name}\' already exists in database.'
                        )
                    elif ExportWritePolicy.REPLACE == if_exists:
                        if drop_table_on_replace:
                            cmd = f'DROP TABLE {full_table_name}'
                            if cascade_on_drop:
                                cmd = f'{cmd} CASCADE'
                            cur.execute(cmd)
                            should_create_table = True
                        else:
                            cur.execute(f'DELETE FROM {full_table_name}')

                if query_string:
                    query = self.build_create_table_as_command(
                        full_table_name,
                        query_string,
                    )

                    if ExportWritePolicy.APPEND == if_exists and table_exists:
                        query = 'INSERT INTO {}\n{}'.format(
                            full_table_name,
                            query_string,
                        )
                    cur.execute(query)
                else:
                    db_dtypes = {col: self.get_type(df[col], dtypes[col]) for col in dtypes}
                    if should_create_table:
                        query = self.build_create_table_command(
                            db_dtypes,
                            schema_name,
                            table_name,
                            unique_constraints=unique_constraints,
                        )
                        cur.execute(query)

                    self.upload_dataframe(
                        cur,
                        df,
                        db_dtypes,
                        dtypes,
                        full_table_name,
                        buffer,
                        allow_reserved_words=allow_reserved_words,
                        unique_conflict_method=unique_conflict_method,
                        unique_constraints=unique_constraints,
                    )
            self.conn.commit()

        if verbose:
            with self.printer.print_msg(
                f'Exporting data to \'{full_table_name}\''
            ):
                __process()
        else:
            __process()

    def clean(self, column: Series, dtype: str) -> Series:
        """
        Cleans column in order to write data frame to PostgreSQL database

        Args:
            column (Series): Column to clean
            dtype (str): The pandas data types of this column

        Returns:
            Series: Cleaned column
        """
        if dtype == PandasTypes.CATEGORICAL:
            return column.astype(str)
        elif dtype in (PandasTypes.TIMEDELTA, PandasTypes.TIMEDELTA64, PandasTypes.PERIOD):
            return column.view(int)
        else:
            return column



from enum import Enum
from typing import Any, Mapping, Union
import pathlib
import os
import yaml


class IOConfigKeys(str, Enum):
    AWS = 'AWS'
    BIGQUERY = 'BigQuery'
    CLICKHOUSE = 'ClickHouse'
    DRUID = 'Druid'
    FILE = 'File'
    POSTGRES = 'PostgreSQL'
    REDSHIFT = 'Redshift'
    S3 = 'S3'
    SNOWFLAKE = 'Snowflake'


class IOConfig:
    """
    Wrapper around IO configuration file.
    """

    def __init__(
        self,
        filepath: Union[os.PathLike, str] = './default_repo/io_config.yaml'
    ) -> None:
        """
        Initializes IO Configuration loader

        Args:
            filepath (os.PathLike): Path to IO configuration file.
        """
        self.filepath = pathlib.Path(filepath)

    def use(self, profile: str) -> Mapping[str, Any]:
        """
        Specifies the profile to use. Profiles are sets of configuration settings.

        Args:
            profile (str): Name of the profile to use.

        Returns:
            Mapping[str, Any]: Configuration settings for this profile
        """
        with self.filepath.open('r') as fin:
            config = yaml.full_load(fin.read())
        profile_settings = config.get(profile)
        if profile_settings is None:
            raise ValueError(f'Invalid configuration profile specified: \'{profile}\'')
        return profile_settings



from mage_ai.api.operations import constants
from mage_ai.api.presenters.BasePresenter import BasePresenter
from mage_ai.shared.hash import extract


class UserPresenter(BasePresenter):
    default_attributes = [
        'avatar',
        'email',
        'first_name',
        'id',
        'last_name',
        'owner',
        'roles',
        'roles_display',
        'roles_new',
        'project_access',
        'username',
    ]

    def present(self, **kwargs):
        data = self.model.to_dict(include_attributes=self.default_attributes)
        data = extract(data, self.default_attributes, include_blank_values=True)

        return data


UserPresenter.register_format(
    constants.CREATE, UserPresenter.default_attributes + ['token', ])



from mage_ai.api.presenters.BasePresenter import BasePresenter


class LogPresenter(BasePresenter):
    default_attributes = [
        'block_run_logs',
        'pipeline_run_logs',
        'total_block_run_log_count',
        'total_pipeline_run_log_count',
    ]



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.errors import FileExistsError
from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db import safe_db_query
from typing import Dict
import urllib.parse


class FileResource(GenericResource):
    @classmethod
    @safe_db_query
    def collection(self, query, meta, user, **kwargs):
        return self.build_result_set(
            [File.get_all_files(get_repo_path())],
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload: Dict, user, **kwargs) -> 'FileResource':
        dir_path = payload['dir_path']
        repo_path = get_repo_path()
        content = None

        if 'file' in payload:
            file = payload['file'][0]
            filename = file['filename']
            content = file['body']
        else:
            filename = payload['name']

        try:
            file = File.create(
                filename,
                dir_path,
                repo_path=repo_path,
                content=content,
                overwrite=payload.get('overwrite', False),
            )

            return self(file, user, **kwargs)
        except FileExistsError as err:
            error = ApiError.RESOURCE_INVALID.copy()
            error.update(dict(message=str(err)))
            raise ApiError(error)

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        file = self.get_model(pk)
        if not file.exists():
            error = ApiError.RESOURCE_NOT_FOUND.copy()
            error.update(message=f'File at {pk} cannot be found.')
            raise ApiError(error)

        return self(file, user, **kwargs)

    @classmethod
    @safe_db_query
    def get_model(self, pk):
        file_path = urllib.parse.unquote(pk)
        return File.from_path(file_path, get_repo_path())

    @safe_db_query
    def delete(self, **kwargs):
        return self.model.delete()

    @safe_db_query
    def update(self, payload, **kwargs):
        self.model.rename(payload['dir_path'], payload['name'])
        return self
